#! /usr/bin/env python
"""This module implements the Atom 1.0 specification defined in RFC 4287

References:

IRIs [RFC3987]; cf URI [RFC3986]
(1) when an IRI that is not also a URI is given for dereferencing,
it MUST be mapped to a URI using the steps in Section 3.1 of [RFC3987]
(2) when an IRI is serving as an atom:id value, it MUST NOT be so mapped,
so that the comparison works as described in Section 4.2.6.1.

xml:base attribute [W3C.REC-xmlbase-20010627]
xml:lang attribute [W3C.REC-xml-20040204], Section 2.12

A Date construct is an element whose content MUST conform to the
"date-time" production in [RFC3339]
"""

import itertools

from . import html401 as html
from . import info
from . import iso8601
from . import rfc2396 as uri
from .pep8 import old_function
from .py2 import is_text, to_text
from .xml import structures as xml
from .xml import namespace as xmlns
from .xml import xsdatatypes as xsi


#: The namespace to use for Atom Document elements
ATOM_NAMESPACE = "http://www.w3.org/2005/Atom"

#: The mime type for Atom Document
ATOM_MIMETYPE = "application/atom+xml"

_ATOM_TEXT_TYPES = {'text': 1, 'html': 1, 'xhtml': 1}


class AtomElement(xmlns.XMLNSElement):

    """Base class for all APP elements.

    All atom elements can have xml:base and xml:lang attributes, these are
    handled by the :py:class:`~pyslet.xml.structures.Element` base
    class.

    See :py:meth:`~pyslet.xml.structures.Element.GetLang` and
    :py:meth:`~pyslet.xml.structures.Element.SetLang`,
    :py:meth:`~pyslet.xml.structures.Element.GetBase` and
    :py:meth:`~pyslet.xml.structures.Element.SetBase`"""
    pass


class TextType(xsi.Enumeration):

    """text type enumeration::

             "text" | "html" | "xhtml"

    This enumeration is used for setting the :py:attr:`Text.type` attribute.

    Usage: TextType.text, TextType.html, TextType.xhtml"""
    decode = {
        'text': 1,
        'html': 2,
        'xhtml': 3
    }


class Text(AtomElement):

    """Base class for atomPlainTextConstruct and atomXHTMLTextConstruct."""

    XMLATTR_type = ('type', TextType.from_str_lower, TextType.to_str)

    def __init__(self, parent):
        AtomElement.__init__(self, parent)
        self.type = TextType.text

    def set_value(self, value, type=TextType.text):
        """Sets the value of the element.  *type* must be a value from
        the :py:class:`TextType` enumeration

        Overloads the basic
        :py:meth:`~pyslet.xml.structures.Element.SetValue`
        implementation, adding an additional *type* attribute to enable
        the value to be set to either a plain TextType.text,
        TextType.html or TextType.xhtml value.  In the case of an xhtml
        type, *value* is parsed for the required XHTML div element and
        this becomes the only child of the element.  Given that the div
        itself is not considered to be part of the content the value can
        be given without the enclosing div, in which case it is
        generated automatically."""
        if type == TextType.text or type == TextType.html:
            AtomElement.set_value(self, value)
            self.type = type
        elif type == TextType.xhtml:
            e = xml.XMLEntity(value)
            doc = html.XHTMLDocument(base_uri=self.resolve_base())
            doc.read_from_entity(e)
            div = list(doc.root.Body.get_children())
            if len(div) == 1 and isinstance(div[0], html.Div):
                div = div[0]
                # We remove our existing content
                self.set_value(None)
                # And do a deep copy of the div instead
                new_div = div.deepcopy(self)
            else:
                new_div = self.add_child(html.Div)
                for divChild in div:
                    if isinstance(divChild, xml.Element):
                        divChild.deepcopy(new_div)
                    else:
                        new_div.add_data(divChild)
            new_div.make_prefix(html.XHTML_NAMESPACE, '')
            self.type = type
        else:
            raise ValueError(
                "Expected text or html identifiers, found %s" % str(type))

    def get_value(self):
        """Gets a single unicode string representing the value of the element.

        Overloads the basic
        :py:meth:`~pyslet.xml.structures.Element.get_value`
        implementation to add support for text of type xhtml.

        When getting the value of TextType.xhtml text the child div element is
        not returned as it is not considered to be part of the content."""
        if self.type == TextType.text or self.type == TextType.html:
            return AtomElement.get_value(self)
        elif self.type == TextType.xhtml:
            # concatenate all children, but should be just a single div
            result = []
            value_children = list(self.get_children())
            if len(value_children) and isinstance(value_children[0], html.Div):
                value_children = list(value_children[0].get_children())
            for c in value_children:
                result.append(to_text(c))
            return ''.join(result)
        else:
            raise ValueError("Unknown text type: %s" % str(self.type))


class AtomId(AtomElement):

    """A permanent, universally unique identifier for an entry or feed."""
    XMLNAME = (ATOM_NAMESPACE, 'id')


class Name(AtomElement):

    """A human-readable name for a person."""
    XMLNAME = (ATOM_NAMESPACE, 'name')


class Title(Text):

    """A :py:class:`Text` construct that conveys a human-readable title
    for an entry or feed."""
    XMLNAME = (ATOM_NAMESPACE, 'title')


class Subtitle(Text):

    """A :py:class:`Text` construct that conveys a human-readable
    description or subtitle for a feed."""
    XMLNAME = (ATOM_NAMESPACE, 'subtitle')


class Summary(Text):

    """A :py:class:`Text` construct that conveys a short summary,
    abstract, or excerpt of an entry."""
    XMLNAME = (ATOM_NAMESPACE, 'summary')


class Rights(Text):

    """A Text construct that conveys information about rights held in
    and over an entry or feed."""
    XMLNAME = (ATOM_NAMESPACE, 'rights')


class Date(AtomElement):

    """An element conforming to the definition of date-time in RFC3339.

    This class is modeled using the iso8601 module."""

    def __init__(self, parent):
        AtomElement.__init__(self, parent)
        #: a :py:class:`~pyslet.iso8601.TimePoint` instance representing
        #: this date
        self.date = iso8601.TimePoint()

    def get_value(self):
        """Overrides
        :py:meth:`~pyslet.xml.structures.Element.get_value`, returning a
        :py:class:`pyslet.iso8601.TimePoint` instance."""
        return self.date

    def set_value(self, value):
        """Overrides :py:meth:`~pyslet.xml.structures.Element.SetValue`,
        enabling the value to be set from a
        :py:class:`pyslet.iso8601.TimePoint` instance.

        If *value* is a string the behaviour is unchanged, if *value* is
        a TimePoint instance then it is formatted using the extended
        format of ISO 8601 in accordance with the requirements of the
        Atom specification."""
        if isinstance(value, iso8601.TimePoint):
            self.date = value
            super(Date, self).set_value(value.get_calendar_string())
        else:
            super(Date, self).set_value(value)
            self.content_changed()

    def content_changed(self):
        """Re-reads the value of the element and sets :py:attr:`date`
        accordingly."""
        self.date = iso8601.TimePoint.from_str(AtomElement.get_value(self))


class Updated(Date):

    """A Date construct indicating the most recent instant in time when
    an entry or feed was modified in a way the publisher considers
    significant."""
    XMLNAME = (ATOM_NAMESPACE, 'updated')


class Published(Date):

    """A Date construct indicating an instant in time associated with an
    event early in the life cycle of the entry."""
    XMLNAME = (ATOM_NAMESPACE, "published")


class Link(AtomElement):

    """A reference from an entry or feed to a Web resource."""
    XMLNAME = (ATOM_NAMESPACE, 'link')
    XMLATTR_href = ('href', uri.URI.from_octets, str)
    XMLATTR_rel = 'rel'
    XMLATTR_type = 'type'
    XMLATTR_hreflang = 'hreflang'
    XMLATTR_title = 'title'
    XMLATTR_length = ('length', xsi.integer_from_str, xsi.integer_to_str)

    def __init__(self, parent):
        AtomElement.__init__(self, parent)
        #: a :py:class:`~pyslet.rfc2396.URI` instance, the link's IRI
        self.href = None
        self.rel = None         #: a string indicating the link relation type
        self.type = None        #: an advisory media type
        #: the language of the resource pointed to by :py:attr:`href`
        self.hreflang = None
        self.title = None       #: human-readable information about the link
        #: an advisory length of the linked content in octets
        self.length = None


class Icon(AtomElement):

    """An image that provides iconic visual identification for a feed."""
    XMLNAME = (ATOM_NAMESPACE, 'icon')

    def __init__(self, parent):
        AtomElement.__init__(self, parent)
        #: a :py:class:`~pyslet.rfc2396.URI` instance representing the
        #: URI of the icon
        self.uri = None

    def get_value(self):
        """Returning a :py:class:`pyslet.rfc2396.URI` instance."""
        return self.uri

    def set_value(self, value):
        """Enables the value to be set from a URI instance.

        If *value* is a string it is used to set the element's content,
        :py:meth:`content_changed` is then called to update the value of
        :py:attr:`uri`.  If *value* is a URI instance then
        :py:attr:`uri` is set directory and it is then converted to a
        string and used to set the element's content."""
        if isinstance(value, uri.URI):
            self.uri = value
            AtomElement.set_value(self, str(value))
        else:
            AtomElement.set_value(self, value)
            self.content_changed()

    def content_changed(self):
        """Sets :py:attr:`uri` accordingly."""
        self.uri = uri.URI.from_octets(AtomElement.get_value(self))


class Logo(Icon):

    """An image that provides visual identification for a feed."""
    XMLNAME = (ATOM_NAMESPACE, 'logo')


class Generator(AtomElement):

    """Identifies the agent used to generate a feed

    The agent is used for debugging and other purposes."""
    XMLNAME = (ATOM_NAMESPACE, 'generator')
    XMLATTR_uri = ('uri', uri.URI.from_octets, str)
    XMLATTR_version = 'version'

    def __init__(self, parent):
        AtomElement.__init__(self, parent)
        #: the uri of the tool used to generate the feed
        self.uri = None
        #: the version of the tool used to generate the feed
        self.version = None

    def set_pyslet_info(self):
        """Sets this generator to a default value

        A representation of this Pyslet module."""
        self.uri = uri.URI.from_octets(info.home)
        self.version = info.version
        self.set_value(info.title)


@old_function('DecodeContentType')
def decode_content_type(src):
    try:
        return TextType.from_str_lower(src)
    except ValueError:
        return src.strip()


@old_function('EncodeContentType')
def encode_content_type(value):
    if is_text(value):
        return value
    else:
        return TextType.to_str(value)


class Content(Text):

    """Contains or links to the content of the entry.

    Although derived from :py:class:`Text` this class overloads the
    meaning of the :py:attr:`Text.type` attribute allowing it to be a
    media type."""
    XMLNAME = (ATOM_NAMESPACE, "content")
    XMLATTR_src = ('src', uri.URI.from_octets, str)
    XMLATTR_type = ('type', decode_content_type, encode_content_type)

    def __init__(self, parent):
        Text.__init__(self, parent)
        self.src = None         #: link to remote content

    def get_value(self):
        """Gets a single string representing the value of the element.

        Overloads the basic :py:meth:`~Text.get_value`, if
        :py:attr:`type` is a media type rather than one of the text
        types then a ValueError is raised."""
        if is_text(self.type):
            raise ValueError("Can't get value of non-text content")
        else:
            Text.get_value(self)


class URI(AtomElement):

    """An IRI associated with a person"""
    XMLNAME = (ATOM_NAMESPACE, 'uri')


class Email(AtomElement):

    """An e-mail address associated with a person"""
    XMLNAME = (ATOM_NAMESPACE, 'email')


class Person(AtomElement):

    """An element that describes a person, corporation, or similar entity"""
    NameClass = Name
    URIClass = URI

    def __init__(self, parent):
        AtomElement.__init__(self, parent)
        self.Name = self.NameClass(self)
        self.URI = None
        self.Email = None

    def get_children(self):
        if self.Name:
            yield self.Name
        if self.URI:
            yield self.URI
        if self.Email:
            yield self.Email
        for child in AtomElement.get_children(self):
            yield child


class Author(Person):

    """A Person construct that indicates the author of the entry or feed."""
    XMLNAME = (ATOM_NAMESPACE, 'author')


class Contributor(Person):

    """A Person construct representing a contributor

    Indicates a person or other entity who contributed to the entry or
    feed."""
    XMLNAME = (ATOM_NAMESPACE, "contributor")


class Category(AtomElement):

    """Information about a category associated with an entry or feed."""
    XMLNAME = (ATOM_NAMESPACE, "category")
    XMLATTR_term = 'term'
    XMLATTR_scheme = 'scheme'
    XMLATTR_label = 'label'

    def __init__(self, parent):
        AtomElement.__init__(self, parent)
        #: a string that identifies the category to which the entry or
        #: feed belongs
        self.term = None
        self.scheme = None
        """an IRI that identifies a categorization scheme.

        This is not converted to a :py:class:`pyslet.rfc2396.URI`
        instance as it is not normally resolved to a resource.  Instead
        it defines a type of namespace."""
        #: a human-readable label for display in end-user applications
        self.label = None


class Entity(AtomElement):

    """Base class for feed, entry and source elements."""

    LinkClass = Link

    def __init__(self, parent):
        AtomElement.__init__(self, parent)
        self.AtomId = None
        """the atomId of the object

        Note that we qualify the class name used to represent the id to
        avoid confusion with the existing 'id' attribute in
        :py:class:`~pyslet.xml.structures.Element`."""
        self.Author = []        #: atomAuthor
        self.Category = []      #: atomCategory
        self.Contributor = []   #: atomContributor
        self.Link = []          #: atomLink
        self.Rights = None      #: atomRights
        self.Title = None       #: atomTitle
        self.Updated = None     #: atomUpdated

    def reset(self):
        for child in itertools.chain(self.Author, self.Category,
                                     self.Contributor, self.Link):
            child.DetachFromParent
        self.Author = []
        self.Category = []
        self.Contributor = []
        self.Link = []
        if self.Rights:
            self.Rights.detach_from_parent()
            self.Rights = None
        if self.Title:
            self.Title.detach_from_parent()
            self.Title = None
        if self.Updated:
            self.Updated.detach_from_parent()
            self.Updated = None
        super(Entity, self).reset()

    def get_children(self):
        if self.AtomId:
            yield self.AtomId
        if self.Title:
            yield self.Title
        if self.Rights:
            yield self.Rights
        if self.Updated:
            yield self.Updated
        for child in itertools.chain(
                self.Link,
                self.Author,
                self.Contributor,
                self.Category,
                AtomElement.get_children(self)):
            yield child


class Source(Entity):

    """Metadata from the original source feed of an entry.

    This class is also used a base class for :py:class:`Feed`."""
    XMLNAME = (ATOM_NAMESPACE, 'source')
    XMLCONTENT = xml.ElementType.ElementContent

    def __init__(self, parent):
        Entity.__init__(self, parent)
        self.Generator = None       #: atomGenerator
        self.Icon = None            #: atomIcon
        self.Logo = None            #: atomLogo
        self.Subtitle = None        #: atomSubtitle

    def get_children(self):
        for child in Entity.get_children(self):
            yield child
        if self.Generator:
            yield self.Generator
        if self.Icon:
            yield self.Icon
        if self.Logo:
            yield self.Logo
        if self.Subtitle:
            yield self.Subtitle


class Feed(Source):

    """Represents an Atom feed.

    This is the document (i.e., top-level) element of an Atom Feed Document,
    acting as a container for metadata and data associated with the feed"""
    XMLNAME = (ATOM_NAMESPACE, 'feed')
    AtomIdClass = AtomId
    TitleClass = Title
    UpdatedClass = Updated

    def __init__(self, parent):
        Source.__init__(self, parent)
        self.AtomId = self.AtomIdClass(self)
        self.Title = self.TitleClass(self)
        self.Updated = self.UpdatedClass(self)
        now = iso8601.TimePoint.from_now_utc()
        self.Updated.set_value(now)
        self.Entry = []     #: atomEntry

    def get_children(self):
        for child in itertools.chain(
                Source.get_children(self),
                self.Entry):
            yield child


class Entry(Entity):

    """Represents an individual entry

    Acts as a container for metadata and data associated with the
    entry."""
    XMLNAME = (ATOM_NAMESPACE, 'entry')
    XMLCONTENT = xml.ElementType.ElementContent
    AtomIdClass = AtomId
    TitleClass = Title
    UpdatedClass = Updated
    LinkClass = Link

    def __init__(self, parent):
        Entity.__init__(self, parent)
        self.AtomId = self.AtomIdClass(self)
        self.Title = self.TitleClass(self)
        self.Updated = self.UpdatedClass(self)
        now = iso8601.TimePoint.from_now_utc()
        self.Updated.set_value(now)
        self.Content = None
        self.Published = None
        self.Source = None
        self.Summary = None

    def reset(self):
        self.AtomId.reset()
        if self.Content:
            self.Content.DetatchFromParent()
            self.Content = None
        if self.Published:
            self.Published.detach_from_parent()
            self.Published = None
        if self.Source:
            self.Source.detach_from_parent()
            self.Source = None
        if self.Summary:
            self.Summary.detach_from_parent()
            self.Summary = None
        super(Entry, self).reset()
        # Parent reset removes 'optional' Title and Updated elements
        self.Title = self.TitleClass(self)
        self.Updated = self.UpdatedClass(self)
        now = iso8601.TimePoint.from_now_utc()
        self.Updated.set_value(now)

    def get_children(self):
        for child in Entity.get_children(self):
            yield child
        if self.Content:
            yield self.Content
        if self.Published:
            yield self.Published
        if self.Source:
            yield self.Source
        if self.Summary:
            yield self.Summary


class AtomDocument(xmlns.XMLNSDocument):
    classMap = {}

    default_ns = ATOM_NAMESPACE

    def __init__(self, **args):
        """"""
        xmlns.XMLNSDocument.__init__(self, defaultNS=ATOM_NAMESPACE, **args)

    @classmethod
    def get_element_class(cls, name):
        return AtomDocument.classMap.get(
            name, AtomDocument.classMap.get((name[0], None),
                                            xmlns.XMLNSElement))

xmlns.map_class_elements(AtomDocument.classMap, globals())



#! /usr/bin/env python

import sys

from . import content
from . import core
from . import metadata
from .. import html401 as html
from .. import rfc2396 as uri
from ..pep8 import old_method
from ..xml import structures as xml
from ..xml import xsdatatypes as xsi


class AssessmentItem(core.QTIElement, core.DeclarationContainer):

    """An assessment item encompasses the information that is presented to a
    candidate and information about how to score the item::

            <xsd:attributeGroup name="assessmentItem.AttrGroup">
                    <xsd:attribute name="identifier" type="string.Type"
                                    use="required"/>
                    <xsd:attribute name="title" type="string.Type"
                                    use="required"/>
                    <xsd:attribute name="label" type="string256.Type"
                                    use="optional"/>
                    <xsd:attribute ref="xml:lang"/>
                    <xsd:attribute name="adaptive" type="boolean.Type"
                                    use="required"/>
                    <xsd:attribute name="timeDependent" type="boolean.Type"
                                    use="required"/>
                    <xsd:attribute name="toolName" type="string256.Type"
                                    use="optional"/>
                    <xsd:attribute name="toolVersion" type="string256.Type"
                                    use="optional"/>
            </xsd:attributeGroup>

            <xsd:group name="assessmentItem.ContentGroup">
                    <xsd:sequence>
                            <xsd:element ref="responseDeclaration"
                                            minOccurs="0"
                                            maxOccurs="unbounded"/>
                            <xsd:element ref="outcomeDeclaration"
                                            minOccurs="0"
                                            maxOccurs="unbounded"/>
                            <xsd:element ref="templateDeclaration"
                                            minOccurs="0"
                                            maxOccurs="unbounded"/>
                            <xsd:element ref="templateProcessing"
                                            minOccurs="0"
                                            maxOccurs="1"/>
                            <xsd:element ref="stylesheet"
                                            minOccurs="0"
                                            maxOccurs="unbounded"/>
                            <xsd:element ref="itemBody"
                                            minOccurs="0"
                                            maxOccurs="1"/>
                            <xsd:element ref="responseProcessing"
                                            minOccurs="0"
                                            maxOccurs="1"/>
                            <xsd:element ref="modalFeedback"
                                            minOccurs="0"
                                            maxOccurs="unbounded"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'assessmentItem')
    XMLATTR_adaptive = ('adaptive', xsi.boolean_from_str, xsi.boolean_to_str)
    XMLATTR_identifier = 'identifier'
    XMLATTR_label = 'label'
    XMLATTR_timeDependent = (
        'timeDependent', xsi.boolean_from_str, xsi.boolean_to_str)
    XMLATTR_title = 'title'
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        core.QTIElement.__init__(self, parent)
        core.DeclarationContainer.__init__(self)
        self.identifier = None
        self.title = None
        self.label = None
        self.adaptive = False
        self.timeDependent = False
        self.ResponseDeclaration = []
        self.OutcomeDeclaration = []
        self.TemplateDeclaration = []
        self.TemplateProcessing = None
        self.StyleSheet = []
        self.ItemBody = None
        self.ResponseProcessing = None
        self.QTIModalFeedback = []
        self.metadata = metadata.QTIMetadata(None)

    def get_children(self):
        for d in self.ResponseDeclaration:
            yield d
        for d in self.OutcomeDeclaration:
            yield d
        for d in self.TemplateDeclaration:
            yield d
        if self.ItemBody:
            yield self.ItemBody
        if self.ResponseProcessing:
            yield self.ResponseProcessing
        for child in self.QTIModalFeedback:
            yield child

    def content_changed(self):
        self.sort_declarations()

    def sort_declarations(self):
        """Sort each of the variable declaration lists so that they are
        in identifier order.  This is not essential but it does help
        ensure that output is predictable. This method is called
        automatically when reading items from XML files."""
        self.ResponseDeclaration.sort()
        self.OutcomeDeclaration.sort()
        self.TemplateDeclaration.sort()

    def render_html(self, item_state, html_parent=None):
        """Renders this item in html, adding nodes to *html_parent*.
        The state of the item (e.g., the values of any controls and
        template variables), is taken from *item_state*, a
        :py:class:`variables.ItemSessionState` instance.

        The result is the top-level div containing the item added to the
        html_parent. If html_parent is None then a parentless div is
        created. If the item has no itemBody then an empty Div is
        returned."""
        if self.ItemBody:
            html_div = self.ItemBody.render_html(
                html_parent, content.HTMLProfile, item_state)
        else:
            if html_parent:
                html_div = html_parent.add_child(html.Div)
            else:
                html_div = html.Div(None)
        return html_div

    @old_method('AddToContentPackage')
    def add_to_content_package(self, cp, lom, dname=None):
        """Adds a resource and associated files to the content package."""
        resource_id = cp.manifest.get_unique_id(self.identifier)
        resource = cp.manifest.root.Resources.add_child(
            cp.manifest.root.Resources.ResourceClass)
        resource.set_id(resource_id)
        resource.type = core.IMSQTI_ITEM_RESOURCETYPE
        resource_metadata = resource.add_child(resource.MetadataClass)
        # resource_metadata.AdoptChild(lom)
        # resource_metadata.AdoptChild(self.metadata.deepcopy())
        lom.deepcopy(resource_metadata)
        self.metadata.deepcopy(resource_metadata)
        # Security alert: we're leaning heavily on ValidateIdentifier assuming
        # it returns a good file name
        fpath = (core.ValidateIdentifier(resource_id) +
                 '.xml').encode(sys.getfilesystemencoding())
        if dname:
            fpath = cp.FilePath(dname, fpath)
        fpath = cp.GetUniqueFile(fpath)
        # This will be the path to the file in the package
        full_path = cp.dPath.join(fpath)
        base = uri.URI.from_virtual_path(full_path)
        if isinstance(self.parent, xml.Document):
            # we are the root so we change the document base
            self.parent.set_base(base)
        else:
            self.set_base(base)
        # Turn this file path into a relative URL in the context of the new
        # resource
        href = resource.relative_uri(base)
        f = cp.File(resource, href)
        resource.SetEntryPoint(f)
        for child in self.get_children():
            if isinstance(child, core.QTIElement):
                child.add_to_cpresource(cp, resource, {})
        return resource


#
# Modal Feedback
#
class QTIModalFeedback(content.FlowContainerMixin, core.QTIElement):

    """Represents the modalFeedback element.

    <xsd:attributeGroup name="modalFeedback.AttrGroup">
            <xsd:attribute name="outcomeIdentifier" type="identifier.Type"
            use="required"/>
            <xsd:attribute name="showHide" type="showHide.Type"
            use="required"/>
            <xsd:attribute name="identifier" type="identifier.Type"
            use="required"/>
            <xsd:attribute name="title" type="string.Type" use="optional"/>
    </xsd:attributeGroup>

    <xsd:group name="modalFeedback.ContentGroup">
            <xsd:sequence>
                    <xsd:group ref="flowStatic.ElementGroup" minOccurs="0"
                    maxOccurs="unbounded"/>
            </xsd:sequence>
    </xsd:group>
    """
    XMLNAME = (core.IMSQTI_NAMESPACE, 'modalFeedback')
    XMLATTR_outcomeIdentifier = (
        'outcomeIdentifier', core.ValidateIdentifier, lambda x: x)
    XMLATTR_showHide = (
        'showHide', core.ShowHide.from_str_lower, core.ShowHide.to_str)
    XMLATTR_identifier = ('identifier', core.ValidateIdentifier, lambda x: x)
    XMLATTR_title = 'title'
    XMLCONTENT = xml.XMLMixedContent

    def __init__(self, parent):
        core.QTIElement.__init__(self, parent)
        self.outcomeIdentifier = None
        self.showHide = None
        self.identifier = None
        self.title = None

    def add_child(self, child_class, name=None):
        if issubclass(child_class, html.FlowMixin):
            return core.QTIElement.add_child(self, child_class, name)
        else:
            # This child cannot go in here
            raise core.QTIValidityError(
                "%s in %s" % (repr(name), self.__class__.__name__))



#! /usr/bin/env python

import itertools

from . import core
from . import content
from .. import html401 as html
from ..xml import structures as xml
from ..xml import xsdatatypes as xsi


class Interaction(content.BodyElement):

    """Interactions allow the candidate to interact with the item. Through an
    interaction, the candidate selects or constructs a response::

            <xsd:attributeGroup name="interaction.AttrGroup">
                    <xsd:attributeGroup ref="bodyElement.AttrGroup"/>
                    <xsd:attribute name="responseIdentifier"
                    type="identifier.Type" use="required"/>
            </xsd:attributeGroup>"""
    XMLATTR_responseIdentifier = (
        'responseIdentifier', core.ValidateIdentifier, lambda x: x)

    def __init__(self, parent):
        content.BodyElement.__init__(self, parent)
        self.responseIdentifier = ''


class InlineInteraction(html.InlineMixin, Interaction):

    """Abstract class for interactions that appear inline."""
    pass


class BlockInteraction(html.BlockMixin, Interaction):

    """An interaction that behaves like a block in the content model. Most
    interactions are of this type::

            <xsd:group name="blockInteraction.ContentGroup">
                    <xsd:sequence>
                            <xsd:element ref="prompt" minOccurs="0"
                            maxOccurs="1"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        Interaction.__init__(self, parent)
        self.Prompt = None

    def get_children(self):
        if self.Prompt:
            yield self.Prompt


class Prompt(content.BodyElement):

    """The prompt used in block interactions
    ::

            <xsd:group name="prompt.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="inlineStatic.ElementGroup"
                            minOccurs="0" maxOccurs="unbounded"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'prompt')
    XMLCONTENT = xml.XMLMixedContent

    def __init__(self, parent):
        content.BodyElement.__init__(self, parent)

    def add_child(self, child_class, name=None):
        if issubclass(child_class, html.InlineMixin):
            return content.BodyElement.add_child(self, child_class, name)
        else:
            # This child cannot go in here
            raise core.QTIValidityError(
                "%s in %s" % (child_class.__name__, self.__class__.__name__))

    def render_html(self, parent, profile, item_state):
        html_p = parent.add_child(html.P)
        html_p.style_class = ["prompt"]
        self.render_html_children(html_p, profile, item_state)


class Choice(content.BodyElement):

    """Many of the interactions involve choosing one or more predefined choices
    ::

            <xsd:attributeGroup name="choice.AttrGroup">
                    <xsd:attributeGroup ref="bodyElement.AttrGroup"/>
                    <xsd:attribute name="identifier" type="identifier.Type"
                    use="required"/>
                    <xsd:attribute name="fixed" type="boolean.Type"
                    use="optional"/>
                    <xsd:attribute name="templateIdentifier"
                    type="identifier.Type" use="optional"/>
                    <xsd:attribute name="showHide" type="showHide.Type"
                    use="optional"/>
            </xsd:attributeGroup>"""
    XMLATTR_identifier = ('identifier', core.ValidateIdentifier, lambda x: x)
    XMLATTR_fixed = ('fixed', xsi.boolean_from_str, xsi.boolean_to_str)
    XMLATTR_templateIdentifier = (
        'templateIdentifier', core.ValidateIdentifier, lambda x: x)
    XMLATTR_showHide = (
        'showHide', core.ShowHide.from_str_lower, core.ShowHide.to_str)

    def __init__(self, parent):
        content.BodyElement.__init__(self, parent)
        self.identifier = ''
        self.fixed = None
        self.templateIdentifier = None
        self.showHide = None


class AssociableChoice(Choice):

    """Other interactions involve associating pairs of predefined choices
    ::

            <xsd:attributeGroup name="associableChoice.AttrGroup">
                    <xsd:attributeGroup ref="choice.AttrGroup"/>
                    <xsd:attribute name="matchGroup" use="optional">
                            <xsd:simpleType>
                                    <xsd:list itemType="identifier.Type"/>
                            </xsd:simpleType>
                    </xsd:attribute>
            </xsd:attributeGroup>"""
    XMLATTR_matchGroup = ('matchGroup', core.ValidateIdentifier, lambda x: x)

    def __init__(self, parent):
        Choice.__init__(self, parent)
        self.matchGroup = []


class ChoiceInteraction(BlockInteraction):

    """The choice interaction presents a set of choices to the candidate. The
    candidate's task is to select one or more of the choices, up to a maximum
    of maxChoices::

            <xsd:attributeGroup name="choiceInteraction.AttrGroup">
                    <xsd:attributeGroup ref="blockInteraction.AttrGroup"/>
                    <xsd:attribute name="shuffle" type="boolean.Type"
                    use="required"/>
                    <xsd:attribute name="maxChoices" type="integer.Type"
                    use="required"/>
                    <xsd:attribute name="minChoices" type="integer.Type"
                    use="optional"/>
            </xsd:attributeGroup>

            <xsd:group name="choiceInteraction.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="blockInteraction.ContentGroup"/>
                            <xsd:element ref="simpleChoice" minOccurs="1"
                            maxOccurs="unbounded"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'choiceInteraction')
    XMLATTR_shuffle = ('shuffle', xsi.boolean_from_str, xsi.boolean_to_str)
    XMLATTR_maxChoices = (
        'maxChoices', xsi.integer_from_str, xsi.integer_to_str)
    XMLATTR_minChoices = (
        'minChoices', xsi.integer_from_str, xsi.integer_to_str)

    def __init__(self, parent):
        BlockInteraction.__init__(self, parent)
        self.shuffle = False
        self.maxChoices = 1
        self.minChoices = None
        self.SimpleChoice = []

    def get_children(self):
        for child in itertools.chain(BlockInteraction.get_children(self),
                                     self.SimpleChoice):
            yield child

    def render_html(self, parent, profile, item_state):
        html_div = parent.add_child(html.Div)
        html_div.style_class = ["choiceInteraction"]
        for child in self.get_children():
            child.render_html(html_div, profile, item_state)


class OrderInteraction(BlockInteraction):

    """In an order interaction the candidate's task is to reorder the choices,
    the order in which the choices are displayed initially is significant::

            <xsd:attributeGroup name="orderInteraction.AttrGroup">
                    <xsd:attributeGroup ref="blockInteraction.AttrGroup"/>
                    <xsd:attribute name="shuffle" type="boolean.Type"
                    use="required"/>
                    <xsd:attribute name="minChoices" type="integer.Type"
                    use="optional"/>
                    <xsd:attribute name="maxChoices" type="integer.Type"
                    use="optional"/>
                    <xsd:attribute name="orientation" type="orientation.Type"
                    use="optional"/>
            </xsd:attributeGroup>

            <xsd:group name="orderInteraction.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="blockInteraction.ContentGroup"/>
                            <xsd:element ref="simpleChoice" minOccurs="1"
                            maxOccurs="unbounded"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'orderInteraction')
    XMLATTR_shuffle = ('shuffle', xsi.boolean_from_str, xsi.boolean_to_str)
    XMLATTR_maxChoices = (
        'maxChoices', xsi.integer_from_str, xsi.integer_to_str)
    XMLATTR_minChoices = (
        'minChoices', xsi.integer_from_str, xsi.integer_to_str)
    XMLATTR_orientation = (
        'orientation',
        core.Orientation.from_str_lower,
        core.Orientation.to_str)

    def __init__(self, parent):
        BlockInteraction.__init__(self, parent)
        self.shuffle = False
        self.maxChoices = 1
        self.minChoices = None
        self.orientation = None
        self.SimpleChoice = []

    def get_children(self):
        for child in itertools.chain(
                BlockInteraction.get_children(self),
                self.SimpleChoice):
            yield child

    def render_html(self, parent, profile, item_state):
        html_div = parent.add_child(html.Div)
        html_div.style_class = ["orderInteraction"]
        for child in self.get_children():
            child.render_html(html_div, profile, item_state)


class SimpleChoice(content.FlowContainerMixin, Choice):

    """A SimpleChoice is a choice that contains flow objects; it must not
    contain any nested interactions::

            <xsd:group name="simpleChoice.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="flowStatic.ElementGroup"
                            minOccurs="0" maxOccurs="unbounded"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'simpleChoice')
    XMLCONTENT = xml.XMLMixedContent

    def add_child(self, child_class, name=None):
        if issubclass(child_class, html.FlowMixin):
            return Choice.add_child(self, child_class, name)
        else:
            # This child cannot go in here
            raise core.QTIValidityError(
                "%s in %s" % (repr(name), self.__class__.__name__))

    def render_html(self, parent, profile, item_state):
        # what type of choice are we?
        interaction = self.find_parent(Interaction)
        if isinstance(interaction, ChoiceInteraction):
            # represented as a DL with DD for each choice
            html_div = parent.add_child(html.Div)
            html_div.style_class = ['simpleChoice']
            html_input = html_div.add_child(html.Input)
            sname = interaction.responseIdentifier + ".SAVED"
            if sname in item_state:
                v = item_state[sname]
            else:
                v = item_state[interaction.responseIdentifier]
            if interaction.maxChoices != 1:
                # we need to be a check-box
                html_input.type = html.InputType.checkbox
                html_input.name = item_state.formPrefix + \
                    interaction.responseIdentifier
                html_input.value = self.identifier
                html_input.checked = (self.identifier in v.value)
            else:
                # we should be a radio button
                html_input.type = html.InputType.radio
                html_input.name = item_state.formPrefix + \
                    interaction.responseIdentifier
                html_input.value = self.identifier
                html_input.checked = (v.value == self.identifier)
            parent = html_div
        elif isinstance(interaction, OrderInteraction):
            # we need to be a pull down menu of rank orderings
            raise NotImplementedError
        self.render_html_children(parent, profile, item_state)


class AssociateInteraction(BlockInteraction):

    """An associate interaction is a blockInteraction that presents candidates
    with a number of choices and allows them to create associations between
    them::

            <xsd:attributeGroup name="associateInteraction.AttrGroup">
                    <xsd:attributeGroup ref="blockInteraction.AttrGroup"/>
                    <xsd:attribute name="shuffle" type="boolean.Type"
                    use="required"/>
                    <xsd:attribute name="maxAssociations" type="integer.Type"
                    use="required"/>
                    <xsd:attribute name="minAssociations" type="integer.Type"
                    use="optional"/>
            </xsd:attributeGroup>

            <xsd:group name="associateInteraction.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="blockInteraction.ContentGroup"/>
                            <xsd:element ref="simpleAssociableChoice"
                            minOccurs="1" maxOccurs="unbounded"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'associateInteraction')
    XMLATTR_shuffle = ('shuffle', xsi.boolean_from_str, xsi.boolean_to_str)
    XMLATTR_maxAssociations = (
        'maxAssociations', xsi.integer_from_str, xsi.integer_to_str)
    XMLATTR_minAssociations = (
        'minAssociations', xsi.integer_from_str, xsi.integer_to_str)

    def __init__(self, parent):
        BlockInteraction.__init__(self, parent)
        self.shuffle = False
        self.maxAssociations = 1
        self.minAssociations = None
        self.SimpleAssociableChoice = []

    def get_children(self):
        for child in itertools.chain(
                BlockInteraction.get_children(self),
                self.SimpleAssociableChoice):
            yield child

    def render_html(self, parent, profile, item_state):
        html_div = parent.add_child(html.Div)
        html_div.style_class = ["associateInteraction"]
        for child in self.get_children():
            child.render_html(html_div, profile, item_state)


class MatchInteraction(BlockInteraction):

    """A match interaction is a blockInteraction that presents candidates with
    two sets of choices and allows them to create associates between pairs of
    choices in the two sets, but not between pairs of choices in the same set::

            <xsd:attributeGroup name="matchInteraction.AttrGroup">
                    <xsd:attributeGroup ref="blockInteraction.AttrGroup"/>
                    <xsd:attribute name="shuffle" type="boolean.Type"
                    use="required"/>
                    <xsd:attribute name="maxAssociations" type="integer.Type"
                    use="required"/>
                    <xsd:attribute name="minAssociations" type="integer.Type"
                    use="optional"/>
            </xsd:attributeGroup>

            <xsd:group name="matchInteraction.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="blockInteraction.ContentGroup"/>
                            <xsd:element ref="simpleMatchSet" minOccurs="2"
                            maxOccurs="2"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'matchInteraction')
    XMLATTR_shuffle = ('shuffle', xsi.boolean_from_str, xsi.boolean_to_str)
    XMLATTR_maxAssociations = (
        'maxAssociations', xsi.integer_from_str, xsi.integer_to_str)
    XMLATTR_minAssociations = (
        'minAssociations', xsi.integer_from_str, xsi.integer_to_str)

    def __init__(self, parent):
        BlockInteraction.__init__(self, parent)
        self.shuffle = False
        self.maxAssociations = 1
        self.minAssociations = None
        self.SimpleMatchSet = []

    def get_children(self):
        for child in itertools.chain(
                BlockInteraction.get_children(self),
                self.SimpleMatchSet):
            yield child

    def render_html(self, parent, profile, item_state):
        html_div = parent.add_child(html.Div)
        html_div.style_class = ["matchInteraction"]
        for child in self.get_children():
            child.render_html(html_div, profile, item_state)


class SimpleAssociableChoice(content.FlowContainerMixin, AssociableChoice):

    """associableChoice is a choice that contains flowStatic objects, it must
    not contain nested interactions::

            <xsd:attributeGroup name="simpleAssociableChoice.AttrGroup">
                    <xsd:attributeGroup ref="associableChoice.AttrGroup"/>
                    <xsd:attribute name="matchMax" type="integer.Type"
                    use="required"/>
                    <xsd:attribute name="matchMin" type="integer.Type"
                    use="optional"/>
            </xsd:attributeGroup>

            <xsd:group name="simpleAssociableChoice.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="flowStatic.ElementGroup"
                            minOccurs="0" maxOccurs="unbounded"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'simpleAssociableChoice')
    XMLATTR_matchMax = ('matchMax', xsi.integer_from_str, xsi.integer_to_str)
    XMLATTR_matchMin = ('matchMin', xsi.integer_from_str, xsi.integer_to_str)
    XMLCONTENT = xml.XMLMixedContent

    def __init__(self, parent):
        AssociableChoice.__init__(self, parent)
        self.matchMax = None
        self.matchMin = None

    def add_child(self, child_class, name=None):
        if issubclass(child_class, html.FlowMixin):
            return Choice.add_child(self, child_class, name)
        else:
            # This child cannot go in here
            raise core.QTIValidityError(
                "%s in %s" % (repr(name), self.__class__.__name__))


class SimpleMatchSet(core.QTIElement):

    """Contains an ordered set of choices for the set
    ::

            <xsd:group name="simpleMatchSet.ContentGroup">
                    <xsd:sequence>
                            <xsd:element ref="simpleAssociableChoice"
                            minOccurs="0" maxOccurs="unbounded"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'simpleMatchSet')
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        core.QTIElement.__init__(self, parent)
        self.SimpleAssociableChoice = []

    def get_children(self):
        for child in self.SimpleAssociableChoice:
            yield child


class GapMatchInteraction(BlockInteraction):

    """A gap match interaction is a blockInteraction that contains a number gaps
    that the candidate can fill from an associated set of choices::

            <xsd:attributeGroup name="gapMatchInteraction.AttrGroup">
                    <xsd:attributeGroup ref="blockInteraction.AttrGroup"/>
                    <xsd:attribute name="shuffle" type="boolean.Type"
                    use="required"/>
            </xsd:attributeGroup>

            <xsd:group name="gapMatchInteraction.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="blockInteraction.ContentGroup"/>
                            <xsd:group ref="gapChoice.ElementGroup"
                            minOccurs="1" maxOccurs="unbounded"/>
                            <xsd:group ref="blockStatic.ElementGroup"
                            minOccurs="1" maxOccurs="unbounded"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'gapMatchInteraction')
    XMLATTR_shuffle = ('shuffle', xsi.boolean_from_str, xsi.boolean_to_str)

    def __init__(self, parent):
        BlockInteraction.__init__(self, parent)
        self.shuffle = False
        self.GapChoice = []

    def get_children(self):
        for child in itertools.chain(
                BlockInteraction.get_children(self),
                self.GapChoice):
            yield child
        for child in content.BodyElement.get_children(self):
            yield child

    def add_child(self, child_class, name=None):
        if issubclass(child_class, (Prompt, GapChoice, html.BlockMixin)):
            return BlockInteraction.add_child(self, child_class, name)
        else:
            # This child cannot go in here
            raise core.QTIValidityError(
                "%s in %s" % (repr(name), self.__class__.__name__))

    def render_html(self, parent, profile, item_state):
        html_div = parent.add_child(html.Div)
        html_div.style_class = ["gapMatchInteraction"]
        for child in self.get_children():
            child.render_html(html_div, profile, item_state)


class Gap(html.InlineMixin, AssociableChoice):

    """A gap is an inline element that must only appear within a
    gapMatchInteraction
    ::

            <xsd:attributeGroup name="gap.AttrGroup">
                    <xsd:attributeGroup ref="associableChoice.AttrGroup"/>
                    <xsd:attribute name="required" type="boolean.Type"
                    use="optional"/>
            </xsd:attributeGroup>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'gap')
    XMLATTR_required = ('required', xsi.boolean_from_str, xsi.boolean_to_str)
    XMLCONTENT = xml.XMLEmpty

    def __init__(self, parent):
        AssociableChoice.__init__(self, parent)
        self.required = False


class GapChoice(AssociableChoice):

    """The choices that are used to fill the gaps in a gapMatchInteraction are
    either simple runs of text or single image objects, both derived from
    gapChoice::

            <xsd:attributeGroup name="gapChoice.AttrGroup">
                    <xsd:attributeGroup ref="associableChoice.AttrGroup"/>
                    <xsd:attribute name="matchMax" type="integer.Type"
                    use="required"/>
                    <xsd:attribute name="matchMin" type="integer.Type"
                    use="optional"/>
            </xsd:attributeGroup>"""
    XMLATTR_matchMax = ('matchMax', xsi.integer_from_str, xsi.integer_to_str)
    XMLATTR_matchMin = ('matchMin', xsi.integer_from_str, xsi.integer_to_str)

    def __init__(self, parent):
        AssociableChoice.__init__(self, parent)
        self.matchMax = None
        self.matchMin = None


class GapText(GapChoice):

    """A simple run of text to be inserted into a gap by the user, may be
    subject to variable value substitution with printedVariable::

            <xsd:group name="gapText.ContentGroup">
                    <xsd:sequence>
                            <xsd:element ref="printedVariable" minOccurs="0"
                            maxOccurs="unbounded"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'gapText')
    XMLCONTENT = xml.XMLMixedContent

    def add_child(self, child_class, name=None):
        if issubclass(child_class, content.PrintedVariable):
            return GapChoice.add_child(self, child_class, name)
        else:
            # This child cannot go in here
            raise core.QTIValidityError(
                "%s in %s" % (child_class.__name__, self.__class__.__name__))


class GapImg(GapChoice):

    """A gap image contains a single image object to be inserted into a gap by
    the candidate::

            <xsd:attributeGroup name="gapImg.AttrGroup">
                    <xsd:attributeGroup ref="gapChoice.AttrGroup"/>
                    <xsd:attribute name="objectLabel" type="string.Type"
                    use="optional"/>
            </xsd:attributeGroup>

            <xsd:group name="gapImg.ContentGroup">
                    <xsd:sequence>
                            <xsd:element ref="object" minOccurs="1"
                            maxOccurs="1"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'gapImg')
    XMLATTR_objectLabel = 'objectLabel'
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        GapChoice.__init__(self, parent)
        self.objectLabel = None
        self.Object = html.Object(self)
        content.fix_html_namespace(self.Object)

    def get_children(self):
        yield self.Object


class InlineChoiceInteraction(InlineInteraction):

    """An inline choice is an inlineInteraction that presents the user with a
    set of choices, each of which is a simple piece of text::

            <xsd:attributeGroup name="inlineChoiceInteraction.AttrGroup">
                    <xsd:attributeGroup ref="inlineInteraction.AttrGroup"/>
                    <xsd:attribute name="shuffle" type="boolean.Type"
                    use="required"/>
                    <xsd:attribute name="required" type="boolean.Type"
                    use="optional"/>
            </xsd:attributeGroup>

            <xsd:group name="inlineChoiceInteraction.ContentGroup">
                    <xsd:sequence>
                            <xsd:element ref="inlineChoice" minOccurs="1"
                            maxOccurs="unbounded"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'inlineChoiceInteraction')
    XMLATTR_shuffle = ('shuffle', xsi.boolean_from_str, xsi.boolean_to_str)
    XMLATTR_required = ('required', xsi.boolean_from_str, xsi.boolean_to_str)
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        InlineInteraction.__init__(self, parent)
        self.shuffle = False
        self.required = False
        self.InlineChoice = []

    def get_children(self):
        for child in self.InlineChoice:
            yield child


class InlineChoice(Choice):

    """A simple run of text to be displayed to the user, may be subject to
    variable value substitution with printedVariable::

            <xsd:group name="inlineChoice.ContentGroup">
                    <xsd:sequence>
                            <xsd:element ref="printedVariable" minOccurs="0"
                            maxOccurs="unbounded"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'inlineChoice')
    XMLCONTENT = xml.XMLMixedContent

    def add_child(self, child_class, name=None):
        if issubclass(child_class, content.PrintedVariable):
            return Choice.add_child(self, child_class, name)
        else:
            # This child cannot go in here
            raise core.QTIValidityError(
                "%s in %s" % (child_class.__name__, self.__class__.__name__))


class StringInteractionMixin:

    """Abstract mix-in class for interactions based on free-text input.  String
    interactions can be bound to numeric response variables, instead of
    strings, if desired::

            <xsd:attributeGroup name="stringInteraction.AttrGroup">
                    <xsd:attribute name="base" type="integer.Type"
                    use="optional"/>
                    <xsd:attribute name="stringIdentifier"
                    type="identifier.Type" use="optional"/>
                    <xsd:attribute name="expectedLength" type="integer.Type"
                    use="optional"/>
                    <xsd:attribute name="patternMask" type="string.Type"
                    use="optional"/>
                    <xsd:attribute name="placeholderText" type="string.Type"
                    use="optional"/>
            </xsd:attributeGroup>"""
    XMLATTR_base = ('base', xsi.integer_from_str, xsi.integer_to_str)
    XMLATTR_stringIdentifier = (
        'stringIdentifier', core.ValidateIdentifier, lambda x: x)
    XMLATTR_expectedLength = (
        'expectedLength', xsi.integer_from_str, xsi.integer_to_str)
    XMLATTR_patternMask = 'patternMask'
    XMLATTR_placeholderText = 'placeholderText'

    def __init__(self):
        self.base = None
        self.stringIdentifier = None
        self.expectedLength = None
        self.patternMask = None
        self.placeholderText = None


class TextEntryInteraction(StringInteractionMixin, InlineInteraction):

    """A textEntry interaction is an inlineInteraction that obtains a simple
    piece of text from the candidate."""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'textEntryInteraction')
    XMLCONTENT = xml.XMLEmpty

    def __init__(self, parent):
        InlineInteraction.__init__(self, parent)
        StringInteractionMixin.__init__(self)


class TextFormat(xsi.EnumerationNoCase):

    """Used to control the format of the text entered by the candidate::

            <xsd:simpleType name="textFormat.Type">
                    <xsd:restriction base="xsd:NMTOKEN">
                            <xsd:enumeration value="plain"/>
                            <xsd:enumeration value="preFormatted"/>
                            <xsd:enumeration value="xhtml"/>
                    </xsd:restriction>
            </xsd:simpleType>

    Defines constants for the above formats.  Usage example::

            TextFormat.plain

    Note that::

            TextFormat.DEFAULT == TextFormat.plain

    For more methods see :py:class:`~pyslet.xml.xsdatatypes.Enumeration`"""
    decode = {
        'plain': 1,
        'preFormatted': 2,
        'xhtml': 3
    }
    aliases = {
        None: 'plain'
    }


class ExtendedTextInteraction(StringInteractionMixin, BlockInteraction):

    """An extended text interaction is a blockInteraction that allows the
    candidate to enter an extended amount of text::

            <xsd:attributeGroup name="extendedTextInteraction.AttrGroup">
                    <xsd:attributeGroup ref="blockInteraction.AttrGroup"/>
                    <xsd:attributeGroup ref="stringInteraction.AttrGroup"/>
                    <xsd:attribute name="maxStrings" type="integer.Type"
                    use="optional"/>
                    <xsd:attribute name="minStrings" type="integer.Type"
                    use="optional"/>
                    <xsd:attribute name="expectedLines" type="integer.Type"
                    use="optional"/>
                    <xsd:attribute name="format" type="textFormat.Type"
                    use="optional"/>
            </xsd:attributeGroup>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'extendedTextInteraction')
    XMLATTR_maxStrings = (
        'maxStrings', xsi.integer_from_str, xsi.integer_to_str)
    XMLATTR_minStrings = (
        'minStrings', xsi.integer_from_str, xsi.integer_to_str)
    XMLATTR_expectedLines = (
        'expectedLines', xsi.integer_from_str, xsi.integer_to_str)
    XMLATTR_format = (
        'format', TextFormat.from_str_lower, TextFormat.to_str)

    def __init__(self, parent):
        BlockInteraction.__init__(self, parent)
        StringInteractionMixin.__init__(self)
        self.maxStrings = None
        self.minStrings = None
        self.expectedLines = None
        self.format = None


class HottextInteraction(BlockInteraction):

    """The hottext interaction presents a set of choices to the candidate
    represented as selectable runs of text embedded within a surrounding
    context, such as a simple passage of text::

            <xsd:attributeGroup name="hottextInteraction.AttrGroup">
                    <xsd:attributeGroup ref="blockInteraction.AttrGroup"/>
                    <xsd:attribute name="maxChoices" type="integer.Type"
                    use="required"/>
                    <xsd:attribute name="minChoices" type="integer.Type"
                    use="optional"/>
            </xsd:attributeGroup>

            <xsd:group name="hottextInteraction.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="blockInteraction.ContentGroup"/>
                            <xsd:group ref="blockStatic.ElementGroup"
                            minOccurs="1" maxOccurs="unbounded"/>
                    </xsd:sequence>
            </xsd:group>
    """
    XMLNAME = (core.IMSQTI_NAMESPACE, 'hottextInteraction')
    XMLATTR_maxChoices = (
        'maxChoices', xsi.integer_from_str, xsi.integer_to_str)
    XMLATTR_minChoices = (
        'minChoices', xsi.integer_from_str, xsi.integer_to_str)

    def __init__(self, parent):
        BlockInteraction.__init__(self, parent)
        self.maxChoices = 1
        self.minChoices = None

    def add_child(self, child_class, name=None):
        if issubclass(child_class, (Prompt, html.BlockMixin)):
            return BlockInteraction.add_child(self, child_class, name)
        else:
            # This child cannot go in here
            raise core.QTIValidityError(
                "%s in %s" % (repr(name), self.__class__.__name__))


class Hottext(html.FlowMixin, Choice):

    """A hottext area is used within the content of an hottextInteraction to
    provide the individual choices::

            <xsd:group name="hottext.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="inlineStatic.ElementGroup"
                            minOccurs="0" maxOccurs="unbounded"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'hottext')
    XMLCONTENT = xml.XMLMixedContent

    def add_child(self, child_class, name=None):
        if issubclass(child_class, html.InlineMixin):
            return Choice.add_child(self, child_class, name)
        else:
            # This child cannot go in here
            raise core.QTIValidityError(
                "%s in %s" % (repr(name), self.__class__.__name__))


class HotspotMixin:

    """Used by graphic interactions involving images with specially defined
    areas or hotspots::

            <xsd:attributeGroup name="hotspot.AttrGroup">
                    <xsd:attribute name="shape" type="shape.Type"
                    use="required"/>
                    <xsd:attribute name="coords" type="coords.Type"
                    use="required"/>
                    <xsd:attribute name="hotspotLabel" type="string256.Type"
                    use="optional"/>
            </xsd:attributeGroup>"""
    XMLATTR_shape = (
        'shape', core.Shape.from_str_lower, core.Shape.to_str)
    XMLATTR_coords = ('coords', html.Coords.from_str, html.Coords.__unicode__)
    XMLATTR_hotspotLabel = 'hotspotLabel'

    def __init__(self):
        self.shape = None
        self.coords = html.Coords()
        self.hotspotLabel = None


class HotspotChoice(HotspotMixin, Choice):

    """A simple choice consisting of a hot-spot
    ::

            <xsd:attributeGroup name="hotspotChoice.AttrGroup">
                    <xsd:attributeGroup ref="choice.AttrGroup"/>
                    <xsd:attributeGroup ref="hotspot.AttrGroup"/>
            </xsd:attributeGroup>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'hotspotChoice')
    XMLCONTENT = xml.XMLEmpty

    def __init__(self, parent):
        Choice.__init__(self, parent)
        HotspotMixin.__init__(self)


class AssociableHotspot(HotspotMixin, AssociableChoice):

    """An associable choice consisting of a hot-spot
    ::

            <xsd:attributeGroup name="associableHotspot.AttrGroup">
                    <xsd:attributeGroup ref="associableChoice.AttrGroup"/>
                    <xsd:attributeGroup ref="hotspot.AttrGroup"/>
                    <xsd:attribute name="matchMax" type="integer.Type"
                    use="required"/>
                    <xsd:attribute name="matchMin" type="integer.Type"
                    use="optional"/>
            </xsd:attributeGroup>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'associableHotspot')
    XMLATTR_matchMax = ('matchMax', xsi.integer_from_str, xsi.integer_to_str)
    XMLATTR_matchMin = ('matchMin', xsi.integer_from_str, xsi.integer_to_str)
    XMLCONTENT = xml.XMLEmpty

    def __init__(self, parent):
        AssociableChoice.__init__(self, parent)
        HotspotMixin.__init__(self)
        self.matchMax = None
        self.matchMin = None


class GraphicInteraction(BlockInteraction):

    """Abstract class for all graphical interactions
    ::

            <xsd:group name="graphicInteraction.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="blockInteraction.ContentGroup"/>
                            <xsd:element ref="object" minOccurs="1"
                            maxOccurs="1"/>
                    </xsd:sequence>
            </xsd:group>"""

    def __init__(self, parent):
        BlockInteraction.__init__(self, parent)
        self.Object = html.Object(self)
        content.fix_html_namespace(self.Object)

    def get_children(self):
        for child in BlockInteraction.get_children(self):
            yield child
        yield self.Object


class HotspotInteraction(GraphicInteraction):

    """A hotspot interaction is a graphical interaction with a corresponding set
    of choices that are defined as areas of the graphic image::

            <xsd:attributeGroup name="hotspotInteraction.AttrGroup">
                    <xsd:attributeGroup ref="graphicInteraction.AttrGroup"/>
                    <xsd:attribute name="maxChoices" type="integer.Type"
                    use="required"/>
                    <xsd:attribute name="minChoices" type="integer.Type"
                    use="optional"/>
            </xsd:attributeGroup>

            <xsd:group name="hotspotInteraction.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="graphicInteraction.ContentGroup"/>
                            <xsd:element ref="hotspotChoice" minOccurs="1"
                            maxOccurs="unbounded"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'hotspotInteraction')
    XMLATTR_maxChoices = (
        'maxChoices', xsi.integer_from_str, xsi.integer_to_str)
    XMLATTR_minChoices = (
        'minChoices', xsi.integer_from_str, xsi.integer_to_str)

    def __init__(self, parent):
        GraphicInteraction.__init__(self, parent)
        self.maxChoices = 1
        self.minChoices = None
        self.HotspotChoice = []

    def get_children(self):
        return itertools.chain(
            GraphicInteraction.get_children(self),
            self.HotspotChoice)


class SelectPointInteraction(GraphicInteraction):

    """Like hotspotInteraction, a select point interaction is a graphic
    interaction. The candidate's task is to select one or more points::

            <xsd:attributeGroup name="selectPointInteraction.AttrGroup">
                    <xsd:attributeGroup ref="graphicInteraction.AttrGroup"/>
                    <xsd:attribute name="maxChoices" type="integer.Type"
                    use="required"/>
                    <xsd:attribute name="minChoices" type="integer.Type"
                    use="optional"/>
            </xsd:attributeGroup>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'selectPointInteraction')
    XMLATTR_maxChoices = (
        'maxChoices', xsi.integer_from_str, xsi.integer_to_str)
    XMLATTR_minChoices = (
        'minChoices', xsi.integer_from_str, xsi.integer_to_str)

    def __init__(self, parent):
        GraphicInteraction.__init__(self, parent)
        self.maxChoices = 1
        self.minChoices = None


class SliderInteraction(BlockInteraction):

    """The slider interaction presents the candidate with a control for
    selecting a numerical value between a lower and upper bound. It must be
    bound to a response variable with single cardinality with a base-type of
    either integer or float::

            <xsd:attributeGroup name="sliderInteraction.AttrGroup">
                    <xsd:attributeGroup ref="blockInteraction.AttrGroup"/>
                    <xsd:attribute name="lowerBound" type="float.Type"
                    use="required"/>
                    <xsd:attribute name="upperBound" type="float.Type"
                    use="required"/>
                    <xsd:attribute name="step" type="integer.Type"
                    use="optional"/>
                    <xsd:attribute name="stepLabel" type="boolean.Type"
                    use="optional"/>
                    <xsd:attribute name="orientation" type="orientation.Type"
                    use="optional"/>
                    <xsd:attribute name="reverse" type="boolean.Type"
                    use="optional"/>
            </xsd:attributeGroup>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'sliderInteraction')
    XMLATTR_lowerBound = ('lowerBound', xsi.float_from_str, xsi.float_to_str)
    XMLATTR_upperBound = ('upperBound', xsi.float_from_str, xsi.float_to_str)
    XMLATTR_step = ('step', xsi.integer_from_str, xsi.integer_to_str)
    XMLATTR_stepLabel = ('stepLabel', xsi.boolean_from_str, xsi.boolean_to_str)
    XMLATTR_orientation = (
        'orientation',
        core.Orientation.from_str_lower,
        core.Orientation.to_str)
    XMLATTR_reverse = ('reverse', xsi.boolean_from_str, xsi.boolean_to_str)

    def __init__(self, parent):
        BlockInteraction.__init__(self, parent)
        self.lowerBound = None
        self.upperBound = None
        self.step = None
        self.stepLabel = False
        self.orientation = None
        self.reverse = None



#! /usr/bin/env python

import random

from . import core
from . import items
from .. import html401 as html
from ..pep8 import MigratedClass, old_method
from ..py2 import range3
from ..xml import structures as xml
from ..xml import xsdatatypes as xsi


class AssessmentTest(core.QTIElement, core.DeclarationContainer):

    """A test is a group of assessmentItems with an associated set of rules that
    determine which of the items the candidate sees, in what order, and in what
    way the candidate interacts with them. The rules describe the valid paths
    through the test, when responses are submitted for response processing and
    when (if at all) feedback is to be given::

            <xsd:attributeGroup name="assessmentTest.AttrGroup">
                    <xsd:attribute name="identifier" type="string.Type"
                                    use="required"/>
                    <xsd:attribute name="title" type="string.Type"
                                    use="required"/>
                    <xsd:attribute name="toolName" type="string256.Type"
                                    use="optional"/>
                    <xsd:attribute name="toolVersion" type="string256.Type"
                                    use="optional"/>
            </xsd:attributeGroup>

            <xsd:group name="assessmentTest.ContentGroup">
                    <xsd:sequence>
                            <xsd:element ref="outcomeDeclaration" minOccurs="0"
                                            maxOccurs="unbounded"/>
                            <xsd:element ref="timeLimits" minOccurs="0"
                                            maxOccurs="1"/>
                            <xsd:element ref="testPart" minOccurs="1"
                                            maxOccurs="unbounded"/>
                            <xsd:element ref="outcomeProcessing" minOccurs="0"
                                            maxOccurs="1"/>
                            <xsd:element ref="testFeedback" minOccurs="0"
                                            maxOccurs="unbounded"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'assessmentTest')
    XMLATTR_identifier = 'identifier'
    XMLATTR_title = 'title'
    XMLATTR_toolName = 'toolName'
    XMLATTR_toolVersion = 'toolVersion'
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        core.QTIElement.__init__(self, parent)
        core.DeclarationContainer.__init__(self)
        self.identifier = None
        self.title = None
        self.toolName = None
        self.toolVersion = None
        self.OutcomeDeclaration = []
        self.TimeLimits = None
        self.TestPart = []
        self.OutcomeProcessing = None
        self.TestFeedback = []
        # : a dictionary of testPart, assessmentSection and assessmentItemRef
        # keyed on identifier
        self.parts = {}

    def get_children(self):
        for d in self.OutcomeDeclaration:
            yield d
        if self.TimeLimits:
            yield self.TimeLimits
        for d in self.TestPart:
            yield d
        if self.OutcomeProcessing:
            yield self.OutcomeProcessing
        for child in self.TestFeedback:
            yield child

    def content_changed(self):
        self.sort_declarations()

    def sort_declarations(self):
        """Sort the outcome declarations so that they are in identifier order.
        This is not essential but it does help ensure that output is
        predictable. This method is called automatically when reading items
        from XML files."""
        self.OutcomeDeclaration.sort()

    @old_method('RegisterPart')
    def register_part(self, part):
        """Registers a testPart, asssessmentSection or assessmentItemRef in
        :py:attr:`parts`."""
        if part.identifier in self.parts:
            raise KeyError("Duplicate identifier: %s" % part.identifier)
        else:
            self.parts[part.identifier] = part

    @old_method('GetPart')
    def get_part(self, identifier):
        """Returns the testPart, assessmentSection or assessmentItemRef with the
        given identifier."""
        return self.parts[identifier]


class NavigationMode(xsi.Enumeration):

    """The navigation mode determines the general paths that the candidate may
    take. A testPart in linear mode restricts the candidate to attempt each
    item in turn. Once the candidate moves on they are not permitted to return.
    A testPart in nonlinear mode removes this restriction - the candidate is
    free to navigate to any item in the test at any time::

            <xsd:simpleType name="navigationMode.Type">
                    <xsd:restriction base="xsd:NMTOKEN">
                            <xsd:enumeration value="linear"/>
                            <xsd:enumeration value="nonlinear"/>
                    </xsd:restriction>
            </xsd:simpleType>

    Defines constants for the above modes.  Usage example::

            NavigationMode.linear

    Note that::

            NavigationMode.DEFAULT == None

    For more methods see :py:class:`~pyslet.xml.xsdatatypes.Enumeration`"""
    decode = {
        'linear': 1,
        'nonlinear': 2
    }


class SubmissionMode(xsi.Enumeration):

    """The submission mode determines when the candidate's responses are
    submitted for response processing. A testPart in individual mode requires
    the candidate to submit their responses on an item-by-item basis. In
    simultaneous mode the candidate's responses are all submitted together at
    the end of the testPart::

            <xsd:simpleType name="submissionMode.Type">
                    <xsd:restriction base="xsd:NMTOKEN">
                            <xsd:enumeration value="individual"/>
                            <xsd:enumeration value="simultaneous"/>
                    </xsd:restriction>
            </xsd:simpleType>

    Defines constants for the above modes.  Usage example::

            SubmissionMode.individual

    Note that::

            SubmissionMode.DEFAULT == None

    For more methods see :py:class:`~pyslet.xml.xsdatatypes.Enumeration`"""
    decode = {
        'individual': 1,
        'simultaneous': 2
    }


class TestPart(core.QTIElement):

    """Each test is divided into one or more parts which may in turn be divided
    into sections, sub-sections, and so on::

            <xsd:attributeGroup name="testPart.AttrGroup">
                    <xsd:attribute name="identifier" type="identifier.Type"
                                    use="required"/>
                    <xsd:attribute name="navigationMode"
                                    type="navigationMode.Type" use="required"/>
                    <xsd:attribute name="submissionMode"
                                    type="submissionMode.Type" use="required"/>
            </xsd:attributeGroup>

            <xsd:group name="testPart.ContentGroup">
                    <xsd:sequence>
                            <xsd:element ref="preCondition" minOccurs="0"
                                            maxOccurs="unbounded"/>
                            <xsd:element ref="branchRule" minOccurs="0"
                                            maxOccurs="unbounded"/>
                            <xsd:element ref="itemSessionControl" minOccurs="0"
                                            maxOccurs="1"/>
                            <xsd:element ref="timeLimits" minOccurs="0"
                                            maxOccurs="1"/>
                            <xsd:element ref="assessmentSection" minOccurs="1"
                                            maxOccurs="unbounded"/>
                            <xsd:element ref="testFeedback" minOccurs="0"
                                            maxOccurs="unbounded"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'testPart')
    XMLATTR_identifier = 'identifier'
    XMLATTR_navigationMode = (
        'navigationMode',
        NavigationMode.from_str,
        NavigationMode.to_str)
    XMLATTR_submissionMode = (
        'submissionMode',
        SubmissionMode.from_str,
        SubmissionMode.to_str)
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        core.QTIElement.__init__(self, parent)
        self.identifier = None
        self.navigationMode = NavigationMode.DEFAULT
        self.submissionMode = SubmissionMode.DEFAULT
        self.PreCondition = []
        self.BranchRule = []
        self.ItemSessionControl = None
        self.TimeLimits = None
        self.AssessmentSection = []
        self.TestFeedback = []

    def get_children(self):
        for c in self.PreCondition:
            yield c
        for c in self.BranchRule:
            yield c
        if self.ItemSessionControl:
            yield self.ItemSessionControl
        if self.TimeLimits:
            yield self.TimeLimits
        for c in self.AssessmentSection:
            yield c
        for c in self.TestFeedback:
            yield c

    def content_changed(self):
        test = self.find_parent(AssessmentTest)
        if test:
            test.RegisterPart(self)

    def check_pre_conditions(self, state):
        """Returns True if this testPart's pre-conditions are satisfied or if
        there are no pre-conditions in effect."""
        for c in self.PreCondition:
            if not c.Evaluate(state):
                return False
        return True

    def get_branch_target(self, state):
        """Returns the identifier of the testPart to branch to, or the
        pre-defined EXIT_TEST identifier.  If there is no branch rule in effect
        then None is returned.  *state* is a
        :py:class:`variables.TestSessionState` instance used to evaluate the
        branch rule expressions."""
        test = self.find_parent(AssessmentTest)
        for r in self.BranchRule:
            if r.Evaluate(state):
                try:
                    if r.target == "EXIT_TEST":
                        return r.target
                    target = test.GetPart(r.target)
                    if not isinstance(target, TestPart):
                        # test parts can only point at other test parts
                        raise core.ProcessingError(
                            "Target of testPart branch rule is not a"
                            " testPart: %s" % r.target)
                    return r.target
                except KeyError:
                    raise core.ProcessingError(
                        "Target of testPart branch rule has not been"
                        " declared: %s" % r.target)


class Selection(core.QTIElement):

    """The selection class specifies the rules used to select the child elements
    of a section for each test session::

            <xsd:attributeGroup name="selection.AttrGroup">
                    <xsd:attribute name="select" type="integer.Type"
                                    use="required"/>
                    <xsd:attribute name="withReplacement" type="boolean.Type"
                                    use="optional"/>
                    <xsd:anyAttribute namespace="##other"/>
            </xsd:attributeGroup>

            <xsd:group name="selection.ContentGroup">
                    <xsd:sequence>
                    <xsd:any namespace="##any" minOccurs="0"
                                maxOccurs="unbounded" processContents="skip"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'selection')
    XMLATTR_select = ('select', xsi.integer_from_str, xsi.integer_to_str)
    XMLATTR_withReplacement = (
        'with_replacement', xsi.boolean_from_str, xsi.boolean_to_str)
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        core.QTIElement.__init__(self, parent)
        self.select = None
        self.with_replacement = False


class Ordering(core.QTIElement):

    """The ordering class specifies the rule used to arrange the child elements
    of a section following selection. If no ordering rule is given we assume
    that the elements are to be ordered in the order in which they are
    defined::

            <xsd:attributeGroup name="ordering.AttrGroup">
                    <xsd:attribute name="shuffle" type="boolean.Type"
                                    use="required"/>
                    <xsd:anyAttribute namespace="##other"/>
            </xsd:attributeGroup>

            <xsd:group name="ordering.ContentGroup">
                    <xsd:sequence>
                    <xsd:any namespace="##any" minOccurs="0"
                                maxOccurs="unbounded" processContents="skip"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'ordering')
    XMLATTR_shuffle = ('shuffle', xsi.boolean_from_str, xsi.boolean_to_str)
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        core.QTIElement.__init__(self, parent)
        self.shuffle = False


class SectionPart(core.QTIElement):

    """Sections group together individual item references and/or sub-sections. A
    number of common parameters are shared by both types of child element::

            <xsd:attributeGroup name="sectionPart.AttrGroup">
                    <xsd:attribute name="identifier" type="identifier.Type"
                                    use="required"/>
                    <xsd:attribute name="required" type="boolean.Type"
                                    use="optional"/>
                    <xsd:attribute name="fixed" type="boolean.Type"
                                    use="optional"/>
            </xsd:attributeGroup>

            <xsd:group name="sectionPart.ContentGroup">
                    <xsd:sequence>
                            <xsd:element ref="preCondition" minOccurs="0"
                                            maxOccurs="unbounded"/>
                            <xsd:element ref="branchRule" minOccurs="0"
                                            maxOccurs="unbounded"/>
                            <xsd:element ref="itemSessionControl" minOccurs="0"
                                            maxOccurs="1"/>
                            <xsd:element ref="timeLimits" minOccurs="0"
                                            maxOccurs="1"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLATTR_identifier = 'identifier'
    XMLATTR_required = ('required', xsi.boolean_from_str, xsi.boolean_to_str)
    XMLATTR_fixed = ('fixed', xsi.boolean_from_str, xsi.boolean_to_str)

    def __init__(self, parent):
        core.QTIElement.__init__(self, parent)
        self.identifier = None
        self.required = False
        self.fixed = False
        self.PreCondition = []
        self.BranchRule = []
        self.ItemSessionControl = None
        self.TimeLimits = None

    def get_children(self):
        for c in self.PreCondition:
            yield c
        for c in self.BranchRule:
            yield c
        if self.ItemSessionControl:
            yield self.ItemSessionControl
        if self.TimeLimits:
            yield self.TimeLimits

    def content_changed(self):
        test = self.find_parent(AssessmentTest)
        if test:
            test.RegisterPart(self)

    def check_pre_conditions(self, state):
        """Returns True if this item or section's pre-conditions are satisfied
        or if there are no pre-conditions in effect."""
        test_part = self.find_parent(TestPart)
        if test_part.navigationMode != NavigationMode.linear:
            return None
        for c in self.PreCondition:
            if not c.Evaluate(state):
                return False
        return True

    def get_branch_target(self, state):
        """Returns the identifier of the next item or section to branch to, or
        one of the pre-defined EXIT_* identifiers.  If there is no branch rule
        in effect then None is returned.  *state* is a
        :py:class:`variables.TestSessionState` instance used to evaluate the
        branch rule expressions."""
        test = self.find_parent(AssessmentTest)
        test_part = self.find_parent(TestPart)
        if test_part.navigationMode != NavigationMode.linear:
            return None
        for r in self.BranchRule:
            if r.Evaluate(state):
                try:
                    if r.target in ("EXIT_SECTION", "EXIT_TESTPART",
                                    "EXIT_TEST"):
                        return r.target
                    target = test.GetPart(r.target)
                    if not isinstance(target, SectionPart):
                        # section parts can only point at other section parts
                        raise core.ProcessingError(
                            "Target of section or item branch rule is not a"
                            " section or item: %s" % r.target)
                    if target.find_parent(TestPart) is not test_part:
                        raise core.ProcessingError(
                            "Target or section or item branch rule is not in"
                            " the same testPart: %s" % r.target)
                    return r.target
                except KeyError:
                    raise core.ProcessingError(
                        "Target of section or item branch rule has not"
                        " been declared: %s" % r.target)


class AssessmentSection(SectionPart):

    """Represents assessmentSection element
    ::

            <xsd:attributeGroup name="assessmentSection.AttrGroup">
                    <xsd:attributeGroup ref="sectionPart.AttrGroup"/>
                    <xsd:attribute name="title" type="string.Type"
                                    use="required"/>
                    <xsd:attribute name="visible" type="boolean.Type"
                                    use="required"/>
                    <xsd:attribute name="keepTogether" type="boolean.Type"
                                    use="optional"/>
            </xsd:attributeGroup>

            <xsd:group name="assessmentSection.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="sectionPart.ContentGroup"/>
                            <xsd:element ref="selection" minOccurs="0"
                                            maxOccurs="1"/>
                            <xsd:element ref="ordering" minOccurs="0"
                                            maxOccurs="1"/>
                            <xsd:element ref="rubricBlock" minOccurs="0"
                                            maxOccurs="unbounded"/>
                            <xsd:group ref="sectionPart.ElementGroup"
                                            minOccurs="0"
                                            maxOccurs="unbounded"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'assessmentSection')
    XMLATTR_title = 'title'
    XMLATTR_visible = ('visible', xsi.boolean_from_str, xsi.boolean_to_str)
    XMLATTR_keepTogether = (
        'keepTogether', xsi.boolean_from_str, xsi.boolean_to_str)
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        SectionPart.__init__(self, parent)
        self.title = None
        self.visible = None
        self.keepTogether = True
        self.Selection = None
        self.Ordering = None
        self.RubricBlock = []
        self.SectionPart = []

    def get_children(self):
        for c in SectionPart.get_children(self):
            yield c
        if self.Selection:
            yield self.Selection
        if self.Ordering:
            yield self.Ordering
        for c in self.RubricBlock:
            yield c
        for c in self.SectionPart:
            yield c


class AssessmentItemRef(SectionPart):

    """Items are incorporated into the test by reference and not by direct
    aggregation::

            <xsd:attributeGroup name="assessmentItemRef.AttrGroup">
                    <xsd:attributeGroup ref="sectionPart.AttrGroup"/>
                    <xsd:attribute name="href" type="uri.Type" use="required"/>
                    <xsd:attribute name="category" use="optional">
                            <xsd:simpleType>
                                    <xsd:list itemType="identifier.Type"/>
                            </xsd:simpleType>
                    </xsd:attribute>
            </xsd:attributeGroup>

            <xsd:group name="assessmentItemRef.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="sectionPart.ContentGroup"/>
                            <xsd:element ref="variableMapping" minOccurs="0"
                                            maxOccurs="unbounded"/>
                            <xsd:element ref="weight" minOccurs="0"
                                            maxOccurs="unbounded"/>
                            <xsd:element ref="templateDefault" minOccurs="0"
                                            maxOccurs="unbounded"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'assessmentItemRef')
    XMLATTR_href = ('href', html.uri.URI.from_octets, html.to_text)
    XMLATTR_category = ('category', None, None, list)
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        SectionPart.__init__(self, parent)
        self.href = None
        self.item = None
        self.category = []
        self.VariableMapping = []
        self.Weight = []
        self.TemplateDefault = []

    def get_children(self):
        for c in SectionPart.get_children(self):
            yield c
        for c in self.VariableMapping:
            yield c
        for c in self.Weight:
            yield c
        for c in self.TemplateDefault:
            yield c

    @old_method('GetItem')
    def get_item(self):
        """Returns the AssessmentItem referred to by this reference."""
        if self.item is None:
            if self.href:
                from . import xml as qtixml
                item_location = self.resolve_uri(self.href)
                doc = qtixml.QTIDocument(baseURI=item_location)
                doc.read()
                if isinstance(doc.root, items.AssessmentItem):
                    self.item = doc.root
        return self.item

    @old_method('SetTemplateDefaults')
    def set_template_defaults(self, item_state, test_state):
        for td in self.TemplateDefault:
            td.Run(item_state, test_state)


class TestForm(MigratedClass):

    """A TestForm is a particular instance of a test, after selection and
    ordering rules have been applied.

    QTI tests can contain selection and ordering rules that enable basic
    variation between instances, or 'forms' of the test.  Selection and
    ordering is not the only source of variation but it provides the basic
    framework for the test.

    The TestForm acts like a (read-only) sequence of component identifiers. The
    identifiers are the identifiers of the test components in the order they
    have been selected.  Identifiers of test parts and sections are included as
    they are legitimate targets of branch rules and may have their own
    pre-conditions, however, the sequence also contains closing identifiers for
    each section and test part.  A closing identifier is the identifier of the
    section or test part preceded by "-".  For example, a simple test with a
    single part and a single section might appear like this::

            [ "", "PartI", "SectionA", "Q1", "Q2", "-SectionA", "-PartI" ]

    Notice that index 0 is always an empty string corresponding to the test
    itself."""

    def __init__(self, test):
        self.test = test		#: the test from which this form was created
        self.components = []  # : the ordered list of identifiers
        #: a mapping from component identifiers to (lists of) indexes into the
        # component list
        self.map = {}
        # Index 0 represents the test itself!
        self.components.append("")
        for part in self.test.TestPart:
            self.components.append(part.identifier)
            # A part always contains all child sections
            for s in part.AssessmentSection:
                self.components.append(s.identifier)
                # no shuffling in test parts, just add a hidden section as a
                # block
                self.components.extend(self.Select(s))
                self.components.append("-" + s.identifier)
            self.components.append("-" + part.identifier)
        for i in range3(len(self.components)):
            id = self.components[i]
            if id in self.map:
                self.map[id].append(i)
            else:
                self.map[id] = [i]

    @old_method('Select')
    def select(self, section, expand_children=True):
        """Runs the selection and ordering rules for *section*.

        It returns a list of identifiers, not including the identifier of the
        section itself."""
        children = section.SectionPart
        if section.Ordering:
            shuffle = section.Ordering.shuffle
        else:
            shuffle = False
        if section.Selection:
            target_size = section.Selection.select
            with_replacement = section.Selection.with_replacement
        else:
            target_size = len(children)
            with_replacement = False
        selection = []
        bag = list(range3(len(children)))
        shuffle_list = []
        # Step 1: make sure we select required children at least once
        for i in range3(len(children)):
            if children[i].required:
                selection.append(i)
                if not with_replacement:
                    bag.remove(i)
        if len(selection) > target_size:
            raise core.SelectionError(
                "#%s contains a selection rule that selects fewer child"
                " elements than the number of required elements" %
                section.identifier)
        # Step 2: top up the selection until we reach the target size
        while len(selection) < target_size:
            if bag:
                i = random.choice(bag)
                selection.append(i)
                if not with_replacement:
                    bag.remove(i)
            else:
                raise core.SelectionError(
                    "Number of children to select in #%s exceeds the number"
                    " of child elements, use with_replacement to resolve" %
                    section.identifier)
        shuffle_list = []
        # Step 3: sort the list to ensure the position of fixed children is
        # honoured
        selection.sort()
        # Step 4: transform to a list of identifiers...
        # replace invisible sections with their contents if we need to
        # split/shuffle them replace floating children with empty slots and put
        # them in the shuffle list
        new_selection = []
        for i in selection:
            child = children[i]
            invisible_section = isinstance(
                child, AssessmentSection) and not child.visible
            if shuffle and not child.fixed:
                # We're shuffling, add a free slot to the selection
                new_selection.append(None)
                if invisible_section and not child.keepTogether:
                    # the grand-children go into the shuffle_list independently
                    # What does a fixed grand-child mean in this situation?
                    # we raise an error at the moment.  Note that we don't
                    # expand the grand children (unless they are also mixed in
                    # from a nested invisible section)
                    for g_child_id in self.Select(child, False):
                        g_child = self.test.GetPart(g_child_id)
                        if g_child.fixed:
                            raise core.SelectionError(
                                "Fixed child of invisible section #%s is"
                                " subject to parent shuffling, use"
                                " keepTogether to resolve" %
                                child.identifier)
                        shuffle_list.append(g_child_id)
                else:
                    # invisible sections with keepTogether go in to the shuffle
                    # list just like items
                    shuffle_list.append(child.identifier)
            else:
                # We're not shuffling or this child is fixed in position
                # (doesn't matter whether visible or not)
                new_selection.append(child.identifier)
                if isinstance(child, AssessmentSection) and expand_children:
                    new_selection = new_selection + self.Select(child, True)
                    new_selection.append("-" + child.identifier)
        selection = new_selection
        if shuffle_list:
            # Step 5: shuffle!
            random.shuffle(shuffle_list)
            # Expanded invisible sections may mean we have more shuffled
            # items than free slots We need to partition the shuffle
            # list into n buckets where n is the number of slots. We
            # choose to put one item in each bucket initially then
            # randomly assign the rest. This gives the expected result
            # in the case where the shuffle list contains one item for
            # each slot.  It also preserves the relative order of fixed
            # items and ensures that adjacent fixed items are not split
            # by a random choice.  Similarly, items fixed at the start
            # or end of the section remain in place
            i = 0
            buckets = []
            for child in selection:
                if child is None:
                    buckets.append([shuffle_list[i]])
                    i += 1
            while i < len(shuffle_list):
                # choose a random bucket
                random.choice(buckets).append(shuffle_list[i])
                i += 1
            # Now splice the buckets into the selection
            for b in buckets:
                # We need to expand any sections that appear in the buckets
                new_bucket = []
                for childID in b:
                    new_bucket.append(childID)
                    child = self.test.GetPart(childID)
                    if isinstance(child, AssessmentSection):
                        new_bucket.extend(self.Select(childID, True))
                        new_bucket.append("-" + childID)
                i = selection.index(None)
                selection[i:i + 1] = new_bucket
        return selection

    def find(self, pname):
        if pname in self.map:
            return self.map[pname]
        else:
            return []

    def index(self, pname):
        return self.components.index(pname)

    def __len__(self):
        return len(self.components)

    def __getitem__(self, index):
        return self.components[index]

    def __setitem__(self, index, value):
        raise TypeError("TestForms are read-only")

    def __delitem__(self, index):
        raise TypeError("TestForms are read-only")

    def __iter__(self):
        return iter(self.components)



#! /usr/bin/env python

import itertools
import logging

from . import core
from . import variables
from ..pep8 import old_method
from ..xml import structures as xml


class ResponseProcessing(core.QTIElement):

    """Response processing is the process by which the Delivery Engine assigns
    outcomes based on the candidate's responses::

            <xsd:attributeGroup name="responseProcessing.AttrGroup">
                    <xsd:attribute name="template" type="uri.Type"
                                    use="optional"/>
                    <xsd:attribute name="templateLocation" type="uri.Type"
                                    use="optional"/>
            </xsd:attributeGroup>

            <xsd:group name="responseProcessing.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="responseRule.ElementGroup"
                                        minOccurs="0" maxOccurs="unbounded"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'responseProcessing')
    XMLATTR_template = 'template'
    XMLATTR_templateLocation = 'templateLocation'
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        core.QTIElement.__init__(self, parent)
        self.template = None
        self.templateLocation = None
        self.ResponseRule = []

    def get_children(self):
        return itertools.chain(
            self.ResponseRule,
            core.QTIElement.get_children(self))

    @old_method('Run')
    def run(self, state):
        """Runs response processing using the values in *state*.

        *	*state* is an :py:class:`~pyslet.qtiv2.variables.ItemSessionState`
                instance."""
        try:
            for r in self.ResponseRule:
                if r.Run(state):
                    break
        except StopProcessing:
            # raised by exitResponse
            pass


class ResponseRule(core.QTIElement):

    """Abstract class to represent all response rules."""

    @old_method('Run')
    def run(self, state):
        """Abstract method to run this rule using the values in *state*."""
        raise NotImplementedError(
            "Unsupported response rule: <%s>" % repr(self.xmlname))


class ResponseCondition(ResponseRule):

    """If the expression given in a responseIf or responseElseIf evaluates to
    true then the sub-rules contained within it are followed and any following
    responseElseIf or responseElse parts are ignored for this response
    condition::

            <xsd:group name="responseCondition.ContentGroup">
                    <xsd:sequence>
                            <xsd:element ref="responseIf" minOccurs="1"
                                            maxOccurs="1"/>
                            <xsd:element ref="responseElseIf" minOccurs="0"
                                            maxOccurs="unbounded"/>
                            <xsd:element ref="responseElse" minOccurs="0"
                                            maxOccurs="1"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'responseCondition')
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        ResponseRule.__init__(self, parent)
        self.ResponseIf = ResponseIf(self)
        self.ResponseElseIf = []
        self.ResponseElse = None

    def get_children(self):
        if self.ResponseIf:
            yield self.ResponseIf
        for child in self.ResponseElseIf:
            yield child
        if self.ResponseElse:
            yield self.ResponseElse

    def run(self, state):
        if self.ResponseIf.Run(state):
            return
        for c in self.ResponseElseIf:
            if c.Run(state):
                return
        if self.ResponseElse:
            self.ResponseElse.Run(state)


class ResponseIf(core.QTIElement):

    """A responseIf part consists of an expression which must have an effective
    baseType of boolean and single cardinality. If the expression is true then
    the sub-rules are processed, otherwise they are skipped (including if the
    expression is NULL)::

            <xsd:group name="responseIf.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="expression.ElementGroup"
                                        minOccurs="1" maxOccurs="1"/>
                            <xsd:group ref="responseRule.ElementGroup"
                                        minOccurs="0" maxOccurs="unbounded"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'responseIf')
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        core.QTIElement.__init__(self, parent)
        self.Expression = None
        self.ResponseRule = []

    def get_children(self):
        if self.Expression:
            yield self.Expression
        for child in self.ResponseRule:
            yield child

    @old_method('Run')
    def run(self, state):
        """Run this test and, if True, any resulting rules.

        Returns *True* if the condition evaluated to *True*."""
        if self.Expression is None:
            raise core.ProcessingError("responseIf with missing condition")
        value = self.Expression.Evaluate(state)
        variables.check_base_types(value.baseType, variables.BaseType.boolean)
        variables.check_cardinalities(
            value.Cardinality(), variables.Cardinality.single)
        if value and value.value:
            for r in self.ResponseRule:
                r.Run(state)
            return True
        else:
            return False


class ResponseElse(core.QTIElement):

    """Represents the responseElse element, see :py:class:`ResponseCondition`
    ::

            <xsd:group name="responseElse.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="responseRule.ElementGroup"
                                        minOccurs="0" maxOccurs="unbounded"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'responseElse')
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        core.QTIElement.__init__(self, parent)
        self.ResponseRule = []

    def get_children(self):
        return iter(self.ResponseRule)

    @old_method('Run')
    def run(self, state):
        """Runs the sub-rules."""
        for r in self.ResponseRule:
            r.Run(state)


class ResponseElseIf(ResponseIf):

    """Represents the responseElse element, see :py:class:`ResponseIf`
    ::

            <xsd:group name="responseElseIf.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="expression.ElementGroup"
                                        minOccurs="1" maxOccurs="1"/>
                            <xsd:group ref="responseRule.ElementGroup"
                                        minOccurs="0" maxOccurs="unbounded"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'responseElseIf')


class SetOutcomeValue(ResponseRule):

    """The setOutcomeValue rule sets the value of an outcome variable to the
    value obtained from the associated expression::

            <xsd:attributeGroup name="setOutcomeValue.AttrGroup">
                    <xsd:attribute name="identifier" type="identifier.Type"
                                    use="required"/>
            </xsd:attributeGroup>

            <xsd:group name="setOutcomeValue.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="expression.ElementGroup"
                                        minOccurs="1" maxOccurs="1"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'setOutcomeValue')
    XMLATTR_identifier = 'identifier'
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        ResponseRule.__init__(self, parent)
        self.identifier = ''
        self.Expression = None

    def get_children(self):
        if self.Expression:
            yield self.Expression

    def run(self, state):
        if self.Expression is None:
            raise core.ProcessingError(
                "setOutcomeValue with missing expression")
        value = self.Expression.Evaluate(state)
        if state.IsOutcome(self.identifier):
            state[self.identifier] = value
        else:
            raise core.ProcessingError(
                "Outcome variable required: %s" % self.identifier)


class StopProcessing(core.QTIError):

    """Raised when a rule which stops processing is encountered."""
    pass


class ExitResponse(ResponseRule):

    """The exit response rule terminates response processing immediately (for
    this invocation).  It does this by raising :py:class:`StopProcessing`::

            <xsd:complexType name="exitResponse.Type"/>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'exitResponse')
    XMLCONTENT = xml.XMLEmpty

    def run(self, state):
        raise StopProcessing


class TemplateProcessing(core.QTIElement):

    """Template processing consists of one or more templateRules that are
    followed by the cloning engine or delivery system in order to assign values
    to the template variables::

            <xsd:group name="templateProcessing.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="templateRule.ElementGroup"
                                        minOccurs="1" maxOccurs="unbounded"/>
                    </xsd:sequence>
            </xsd:group>

            <xsd:complexType name="templateProcessing.Type" mixed="false">
                    <xsd:group ref="templateProcessing.ContentGroup"/>
            </xsd:complexType>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'templateProcessing')
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        core.QTIElement.__init__(self, parent)
        self.TemplateRule = []

    def get_children(self):
        return itertools.chain(
            self.TemplateRule,
            core.QTIElement.get_children(self))

    @old_method('Run')
    def run(self, state):
        """Runs template processing rules using the values in *state*.

        *	*state* is an :py:class:`~pyslet.qtiv2.variables.ItemSessionState`
                instance."""
        try:
            for r in self.TemplateRule:
                if r.Run(state):
                    break
        except StopProcessing:
            # raised by exitTemplate
            pass


class TemplateRule(core.QTIElement):

    """Abstract class to represent all template rules."""

    @old_method('Run')
    def run(self, state):
        """Abstract method to run this rule using the values in *state*."""
        raise NotImplementedError(
            "Unsupported template rule: <%s>" % repr(self.xmlname))


class TemplateCondition(TemplateRule):

    """If the expression given in the templateIf or templateElseIf evaluates to
    true then the sub-rules contained within it are followed and any following
    templateElseIf or templateElse parts are ignored for this template
    condition::

            <xsd:group name="templateCondition.ContentGroup">
                    <xsd:sequence>
                            <xsd:element ref="templateIf" minOccurs="1"
                                            maxOccurs="1"/>
                            <xsd:element ref="templateElseIf" minOccurs="0"
                                            maxOccurs="unbounded"/>
                            <xsd:element ref="templateElse" minOccurs="0"
                                            maxOccurs="1"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'templateCondition')
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        TemplateRule.__init__(self, parent)
        self.TemplateIf = TemplateIf(self)
        self.TemplateElseIf = []
        self.TemplateElse = None

    def get_children(self):
        if self.TemplateIf:
            yield self.TemplateIf
        for child in self.TemplateElseIf:
            yield child
        if self.TemplateElse:
            yield self.TemplateElse

    def run(self, state):
        if self.TemplateIf.Run(state):
            return
        for c in self.TemplateElseIf:
            if c.Run(state):
                return
        if self.TemplateElse:
            self.TemplateElse.Run(state)


class TemplateIf(core.QTIElement):

    """A templateIf part consists of an expression which must have an effective
    baseType of boolean and single cardinality. If the expression is true then
    the sub-rules are processed, otherwise they are skipped (including if the
    expression is NULL)::

            <xsd:group name="templateIf.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="expression.ElementGroup"
                                        minOccurs="1" maxOccurs="1"/>
                            <xsd:group ref="templateRule.ElementGroup"
                                        minOccurs="0" maxOccurs="unbounded"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'templateIf')
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        core.QTIElement.__init__(self, parent)
        self.Expression = None
        self.TemplateRule = []

    def get_children(self):
        if self.Expression:
            yield self.Expression
        for child in self.TemplateRule:
            yield child

    @old_method('Run')
    def run(self, state):
        """Run this test and, if True, any resulting rules.

        Returns *True* if the condition evaluated to *True*."""
        if self.Expression is None:
            raise core.ProcessingError("templateIf with missing condition")
        value = self.Expression.Evaluate(state)
        variables.check_base_types(value.baseType, variables.BaseType.boolean)
        variables.check_cardinalities(
            value.Cardinality(), variables.Cardinality.single)
        if value and value.value:
            for r in self.TemplateRule:
                r.Run(state)
            return True
        else:
            return False


class TemplateElse(core.QTIElement):

    """Represents the templateElse element, see :py:class:`TemplateCondition`
    ::

            <xsd:group name="templateElse.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="templateRule.ElementGroup"
                                        minOccurs="0" maxOccurs="unbounded"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'templateElse')
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        core.QTIElement.__init__(self, parent)
        self.TemplateRule = []

    def get_children(self):
        return iter(self.TemplateRule)

    @old_method('Run')
    def run(self, state):
        """Runs the sub-rules."""
        for r in self.TemplateRule:
            r.Run(state)


class TemplateElseIf(TemplateIf):

    """Represents the templateElse element, see :py:class:`templateIf`
    ::

            <xsd:group name="templateElseIf.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="expression.ElementGroup"
                                        minOccurs="1" maxOccurs="1"/>
                            <xsd:group ref="templateRule.ElementGroup"
                                        minOccurs="0" maxOccurs="unbounded"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'templateElseIf')


class SetTemplateValue(TemplateRule):

    """The setTemplateValue rule sets the value of a template variable to the
    value obtained from the associated expression::

            <xsd:attributeGroup name="setTemplateValue.AttrGroup">
                    <xsd:attribute name="identifier" type="identifier.Type"
                                    use="required"/>
            </xsd:attributeGroup>

            <xsd:group name="setTemplateValue.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="expression.ElementGroup"
                                        minOccurs="1" maxOccurs="1"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'setTemplateValue')
    XMLATTR_identifier = 'identifier'
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        TemplateRule.__init__(self, parent)
        self.identifier = ''
        self.Expression = None

    def get_children(self):
        if self.Expression:
            yield self.Expression

    def run(self, state):
        if self.Expression is None:
            raise core.ProcessingError(
                "setTemplateValue with missing expression")
        value = self.Expression.Evaluate(state)
        if state.IsTemplate(self.identifier):
            state[self.identifier] = value
        else:
            raise core.ProcessingError(
                "Template variable required: %s" % self.identifier)


class SetCorrectResponse(TemplateRule):

    """The setCorrectResponse rule sets the correct value of a response
    variable to the value obtained from the associated expression::

            <xsd:attributeGroup name="setCorrectResponse.AttrGroup">
                    <xsd:attribute name="identifier" type="identifier.Type"
                                    use="required"/>
            </xsd:attributeGroup>

            <xsd:group name="setCorrectResponse.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="expression.ElementGroup"
                                        minOccurs="1" maxOccurs="1"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'setCorrectResponse')
    XMLATTR_identifier = 'identifier'
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        TemplateRule.__init__(self, parent)
        self.identifier = ''
        self.Expression = None

    def get_children(self):
        if self.Expression:
            yield self.Expression

    def run(self, state):
        if self.Expression is None:
            raise core.ProcessingError(
                "setCorrectResponse with missing expression")
        value = self.Expression.Evaluate(state)
        d = state.GetDeclaration(self.identifier)
        if isinstance(d, variables.ResponseDeclaration):
            state[self.identifier + ".CORRECT"] = value
        elif state.IsResponse(self.identifier):
            raise core.ProcessingError(
                "Can't set the correct value of a built-in response %s" %
                self.identifier)
        else:
            raise core.ProcessingError(
                "%s is not a response variable" % self.identifier)


class SetDefaultValue(TemplateRule):

    """The setDefaultValue rule sets the default value of a response or outcome
    variable to the value obtained from the associated expression::

            <xsd:attributeGroup name="setDefaultValue.AttrGroup">
                    <xsd:attribute name="identifier" type="identifier.Type"
                                    use="required"/>
            </xsd:attributeGroup>

            <xsd:group name="setDefaultValue.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="expression.ElementGroup"
                                        minOccurs="1" maxOccurs="1"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'setDefaultValue')
    XMLATTR_identifier = 'identifier'
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        TemplateRule.__init__(self, parent)
        self.identifier = ''
        self.Expression = None

    def get_children(self):
        if self.Expression:
            yield self.Expression

    def run(self, state):
        if self.Expression is None:
            raise core.ProcessingError(
                "setDefaultValue with missing expression")
        value = self.Expression.Evaluate(state)
        d = state.GetDeclaration(self.identifier)
        if (isinstance(d, variables.ResponseDeclaration) or
                isinstance(d, variables.OutcomeDeclaration)):
            state[self.identifier + ".DEFAULT"] = value
        elif (state.IsResponse(self.identifier) or
              state.IsOutcome(self.identifier)):
            raise core.ProcessingError(
                "Can't set the correct value of a built-in variable %s" %
                self.identifier)
        else:
            raise core.ProcessingError(
                "%s is not a response or outcome variable" % self.identifier)


class ExitTemplate(TemplateRule):

    """The exit template rule terminates template processing immediately.  It
    does this by raising :py:class:`StopProcessing`::

            <xsd:complexType name="exitTemplate.Type"/>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'exitTemplate')
    XMLCONTENT = xml.XMLEmpty

    def run(self, state):
        raise StopProcessing


class TestPartCondition(core.QTIElement):

    def __init__(self, parent):
        core.QTIElement.__init__(self, parent)
        self.Expression = None

    def get_children(self):
        if self.Expression:
            yield self.Expression
        for child in core.QTIElement.get_children(self):
            yield child

    @old_method('Evaluate')
    def evaluate(self, state):
        """Evaluates the condition using the values in *state*.

        *	*state* is a :py:class:`~pyslet.qtiv2.variables.TestSessionState`
                instance."""
        if self.Expression is None:
            raise core.ProcessingError(
                "preCondition or branchRule with missing condition")
        value = self.Expression.Evaluate(state)
        variables.check_base_types(value.baseType, variables.BaseType.boolean)
        variables.check_cardinalities(
            value.Cardinality(), variables.Cardinality.single)
        return value and value.value


class PreCondition(TestPartCondition):

    """A preCondition is a simple expression attached to an assessmentSection or
    assessmentItemRef that must evaluate to true if the item is to be
    presented::

            <xsd:group name="preCondition.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="expression.ElementGroup"
                                        minOccurs="1" maxOccurs="1"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'preCondition')
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        TestPartCondition.__init__(self, parent)


class BranchRule(TestPartCondition):

    """A branch-rule is a simple expression attached to an assessmentItemRef,
    assessmentSection or testPart that is evaluated after the item, section, or
    part has been presented to the candidate::

            <xsd:attributeGroup name="branchRule.AttrGroup">
                    <xsd:attribute name="target" type="identifier.Type"
                                    use="required"/>
            </xsd:attributeGroup>

            <xsd:group name="branchRule.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="expression.ElementGroup"
                                        minOccurs="1" maxOccurs="1"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'branchRule')
    XMLATTR_target = 'target'
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        TestPartCondition.__init__(self, parent)
        self.target = None


class TemplateDefault(core.QTIElement):

    """Overrides the default value of a template variable based on the test
    context in which the template is instantiated::

            <xsd:attributeGroup name="templateDefault.AttrGroup">
                    <xsd:attribute name="templateIdentifier"
                                    type="identifier.Type" use="required"/>
            </xsd:attributeGroup>

            <xsd:group name="templateDefault.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="expression.ElementGroup"
                                        minOccurs="1" maxOccurs="1"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'templateDefault')
    XMLATTR_templateIdentifier = 'templateIdentifier'
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        core.QTIElement.__init__(self, parent)
        self.templateIdentifier = None
        self.Expression = None

    def get_children(self):
        if self.Expression:
            yield self.Expression
        for child in core.QTIElement.get_children(self):
            yield child

    @old_method('Run')
    def run(self, item_state, test_state):
        """Updates the value of a template variable in *item_state* based on the
        values in *test_state*."""
        if self.Expression is None:
            raise core.ProcessingError(
                "templateDefault with missing expression")
        value = self.Expression.Evaluate(test_state)
        if self.templateIdentifier is None:
            logging.debug("TemplateDefault.run: %s", self)
        try:
            d = item_state.GetDeclaration(self.templateIdentifier)
        except KeyError:
            raise core.ProcessingError(
                "%s is not a variable" % self.templateIdentifier)
        if isinstance(d, variables.TemplateDeclaration):
            # we don't actually use the .DEFAULT form for template variables as
            # they have their values set directly.
            item_state[self.templateIdentifier] = value
        else:
            raise core.ProcessingError(
                "%s is not a template variable" % self.templateIdentifier)



#! /usr/bin/env python
"""The module creates some basic constants to describe the Pyslet package."""

title_name = u"Pyslet"
name = "pyslet"
copyright = u"\xA92008-2017, Steve Lay"

major_version = "0.7"
build_date = "20170805"
version = "%s.%s" % (major_version, build_date)

title = (
    "Pyslet: "
    "Python package for Standards in Learning, Education and Training")

home = "http://www.pyslet.org/"
